import csv
import functools
import logging
import os

from collections import defaultdict
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, TextIO

import sklearn.pipeline

from sklearn.base import BaseEstimator, TransformerMixin

from bob.bio.base.pipelines.abstract_classes import Database
from bob.pipelines import (
    DelayedSample,
    FileListDatabase,
    Sample,
    SampleSet,
    check_parameters_for_validity,
)
from bob.pipelines.dataset import open_definition_file

from ..utils.annotations import read_annotation_file

logger = logging.getLogger(__name__)


def _sample_sets_to_samples(sample_sets):
    return functools.reduce(
        lambda x, y: x + y, (s.samples for s in sample_sets), []
    )


def _add_key(samples: list[Sample]) -> list[Sample]:
    """Adds a ``key`` attribute to all samples if ``key`` is not present

    Will use ``path`` to create a unique ``key``.
    Note that this won't create unique keys if you have multiple times the same path in
    different samples. This will be problematic, as key are expected to be unique.
    """

    out = []
    for sample in samples:
        if isinstance(sample, SampleSet):
            out.append(SampleSet(samples=_add_key(sample), parent=sample))
            continue
        if not hasattr(sample, "key"):
            if hasattr(sample, "path"):
                sample.key = sample.path
            else:
                raise ValueError(
                    f"Sample has no 'key' and no 'path' to infer it. {sample=}"
                )
        out.append(sample)
    return out


def validate_bio_samples(samples):
    """Validates Samples or SampleSets for backwards compatibility reasons.

    This will add a ``key`` attribute (if not already present) to each sample, copied
    from the path.
    """

    for sample in samples:
        if isinstance(sample, SampleSet):
            validate_bio_samples(sample.samples)
            if not hasattr(sample, "template_id"):
                raise ValueError(
                    f"SampleSet must have a template_id attribute, got {sample}"
                )
            if not hasattr(sample, "subject_id"):
                raise ValueError(
                    f"SampleSet must have a subject_id attribute, got {sample}"
                )
            continue

        if not hasattr(sample, "key"):
            if hasattr(sample, "path"):
                sample.key = sample.path
            else:
                raise ValueError(
                    f"Sample must have a key or a path attribute, got {sample}"
                )

        if not hasattr(sample, "subject_id"):
            raise ValueError(
                f"Sample must have a subject_id attribute, got {sample}"
            )


class CSVDatabase(FileListDatabase, Database):
    """A csv file database.

    The database protocol files should provide the following files:


    .. code-block:: text

        dataset_protocols_path/
        dataset_protocols_path/my_protocol/train/for_background_model.csv
        dataset_protocols_path/my_protocol/train/for_znorm.csv
        dataset_protocols_path/my_protocol/train/for_tnorm.csv
        dataset_protocols_path/my_protocol/dev/for_enrolling.csv
        dataset_protocols_path/my_protocol/dev/for_probing.csv
        dataset_protocols_path/my_protocol/dev/for_matching.csv
        dataset_protocols_path/my_protocol/eval/for_enrolling.csv
        dataset_protocols_path/my_protocol/eval/for_probing.csv
        dataset_protocols_path/my_protocol/eval/for_matching.csv ...

    The ``for_background_model`` file should contain the following columns::

        key,subject_id
        subject1_image1.png,1
        subject1_image2.png,1
        subject2_image1.png,2
        subject2_image2.png,2

    In all the csv files, you can have a column called ``path`` which will be
    used as ``key`` if the ``key`` is not specified. For example::

        path,subject_id
        subject1_image1.png,1
        subject1_image2.png,1
        subject2_image1.png,2
        subject2_image2.png,2

    or::

        path,subject_id,key
        subject1_audio1.wav,1,subject1_audio1_channel1
        subject1_audio1.wav,1,subject1_audio1_channel2
        subject1_audio2.wav,1,subject1_audio2_channel1
        subject1_audio2.wav,1,subject1_audio2_channel2

    The ``key`` column will be used to checkpoint each sample into a unique file and
    must therefore be unique across the whole dataset.

    The ``for_enrolling.csv`` file should contain the following columns::

        key,subject_id,template_id
        subject3_image1.png,3,template_1
        subject3_image2.png,3,template_1
        subject3_image3.png,3,template_2
        subject3_image4.png,3,template_2
        subject4_image1.png,4,template_3
        subject4_image2.png,4,template_3
        subject4_image3.png,4,template_4
        subject4_image4.png,4,template_4

    The ``for_probing.csv`` file should contain the following columns::

        key,subject_id,template_id
        subject5_image1.png,5,template_5
        subject5_image2.png,5,template_5
        subject5_image3.png,5,template_6
        subject5_image4.png,5,template_6
        subject6_image1.png,6,template_7
        subject6_image2.png,6,template_7
        subject6_image3.png,6,template_8
        subject6_image4.png,6,template_8

    Subject identity (``subject_id``) is a unique identifier for one identity (one
    person).
    Template Identity (``template_id``) is an identifier used to group samples when
    they need to be enrolled or scored together.
    :class:`~bob.bio.base.pipelines.BioAlgorithm` will process
    these template.


    By default, each enroll ``template_id`` will be compared against each
    probe ``template_id`` to produce one score per pair. If you want to specify exact
    comparisons (sparse scoring), you can add the ``for_matching.csv`` with the
    following columns::

        enroll_template_id,probe_template_id
        template_1,template_5
        template_2,template_6
        template_3,template_5
        template_3,template_7
        template_4,template_5
        template_4,template_8

    ``for_znorm.csv`` and ``for_tnorm.csv`` files are optional and are used for score
    normalization. See :class:`~bob.bio.base.pipelines.PipelineScoreNorm`.
    ``for_znorm.csv`` has the same format as ``for_probing.csv`` and
    ``for_tnorm.csv`` has the same format as ``for_enrolling.csv``.
    """

    def __init__(
        self,
        *,
        name: str,
        protocol: str,
        dataset_protocols_path: Optional[str] = None,
        transformer: Optional[sklearn.pipeline.Pipeline] = None,
        templates_metadata: Optional[list[str]] = None,
        annotation_type: Optional[str] = None,
        fixed_positions: Optional[dict[str, tuple[float, float]]] = None,
        memory_demanding=False,
        **kwargs,
    ):
        """
        Parameters
        ----------
        name
            The name of the database.
        protocol
            Name of the protocol folder in the CSV definition structure.
        dataset_protocol_path
            Path to the CSV files structure (see :ref:`bob.bio.base.database_interface`
            for more info).
        transformer
            An sklearn pipeline or equivalent transformer that handles some light
            preprocessing of the samples (This will always run locally).
        templates_metadata
            Metadata that originate from the samples and must be present in the
            templates (SampleSet) e.g. ``["gender", "age"]``. This should be metadata
            that is common to all the samples in a template.
        annotation_type
            A string describing the annotations passed to the annotation loading
            function
        fixed_positions
            TODO Why is it here? What does it do exactly?
            --> move it when the FaceCrop annotator is implemented correctly.
        memory_demanding
            Flag that indicates that experiments using this should not run on low-mem
            workers.
        """
        if not hasattr(self, "name"):
            self.name = name
        transformer = sklearn.pipeline.make_pipeline(
            sklearn.pipeline.FunctionTransformer(_add_key), transformer
        )
        super().__init__(
            name=name,  # For FileListDatabase
            protocol=protocol,
            dataset_protocols_path=dataset_protocols_path,
            transformer=transformer,
            annotation_type=annotation_type,
            fixed_positions=fixed_positions,
            memory_demanding=memory_demanding,
            **kwargs,
        )
        if self.list_file("dev", "for_matching") is None:
            self.score_all_vs_all = True
        else:
            self.score_all_vs_all = False

        self.templates_metadata = []
        if templates_metadata is not None:
            self.templates_metadata = templates_metadata

    def list_file(self, group: str, name: str) -> TextIO:
        """Returns a definition file containing one sample per row.

        Overloads ``bob.pipelines`` list_file as the group is a dir.
        """

        try:
            list_file = open_definition_file(
                search_pattern=Path(group) / (name + ".csv"),
                database_name=self.name,
                protocol=self.protocol,
                database_filename=self.dataset_protocols_path.name,
                base_dir=self.dataset_protocols_path.parent,
                subdir=".",
            )
            return list_file
        except FileNotFoundError:
            return None

    def get_reader(self, group: str, name: str) -> Iterable:
        """Returns an :any:`Iterable` containing :class:`Sample` or :class:`SampleSet` objects."""
        key = (self.protocol, group, name)
        if key not in self.readers:
            list_file = self.list_file(group, name)
            self.readers[key] = None
            if list_file is not None:
                self.readers[key] = self.reader_cls(
                    list_file=list_file, transformer=self.transformer
                )

        reader = self.readers[key]
        return reader

    # cached methods should be based on protocol as well
    @functools.lru_cache(maxsize=None)
    def _background_model_samples(self, protocol):
        reader = self.get_reader("train", "for_background_model")
        if reader is None:
            return []
        samples = list(reader)
        validate_bio_samples(samples)
        return samples

    def background_model_samples(self):
        return self._background_model_samples(self.protocol)

    def _sample_sets(self, group, name):
        # we need protocol as input so we can cache the result
        reader = self.get_reader(group, name)
        if reader is None:
            return []
        # create Sample_sets from samples given their unique enroll_template_id/probe_template_id
        samples_grouped_by_template_id = defaultdict(list)
        for sample in reader:
            samples_grouped_by_template_id[sample.template_id].append(sample)
        sample_sets = []
        for (
            template_id,
            samples_for_template_id,
        ) in samples_grouped_by_template_id.items():
            # since all samples inside one sampleset have the same subject_id,
            # we add that as well.
            samples = list(samples_for_template_id)
            subject_id = samples[0].subject_id
            metadata = {
                m: getattr(samples[0], m) for m in self.templates_metadata
            }
            sample_sets.append(
                SampleSet(
                    samples,
                    template_id=template_id,
                    subject_id=subject_id,
                    key=f"template_{template_id}",
                    **metadata,
                )
            )
        validate_bio_samples(sample_sets)
        return sample_sets

    # cached methods should be based on protocol as well
    @functools.lru_cache(maxsize=None)
    def _references(self, protocol, group):  # TODO: protocol
        return self._sample_sets(group, "for_enrolling")

    def references(self, group="dev"):
        return self._references(self.protocol, group)

    def _add_all_references(self, sample_sets, group):
        references = [s.template_id for s in self.references(group)]
        for sample_set in sample_sets:
            sample_set.references = references

    # cached methods should be based on protocol as well
    @functools.lru_cache(maxsize=None)
    def _probes(self, protocol, group):  # TODO: protocol
        sample_sets = self._sample_sets(group, "for_probing")

        # if there are no probes
        if not sample_sets:
            return sample_sets

        # populate .references for each sample set
        matching_file = self.list_file(group, "for_matching")
        if matching_file is None:
            self._add_all_references(sample_sets, group)

        # read the matching file
        else:
            # references is dict where key is probe_template_id and value is a
            # list of enroll_template_ids
            references = defaultdict(list)
            reader = csv.DictReader(matching_file)
            for row in reader:
                references[row["probe_template_id"]].append(
                    row["enroll_template_id"]
                )

            for sample_set in sample_sets:
                sample_set.references = references[sample_set.template_id]

        return sample_sets

    def probes(self, group="dev"):
        return self._probes(self.protocol, group)

    def samples(self, groups=None):
        """Get samples of a certain group

        Parameters
        ----------
        groups : :obj:`str`, optional
            A str or list of str to be used for filtering samples, by default None

        Returns
        -------
        list
            A list containing the samples loaded from csv files.
        """
        groups = check_parameters_for_validity(
            groups, "groups", self.groups(), self.groups()
        )
        all_samples = []

        if "train" in groups:
            all_samples.extend(self.background_model_samples())
            groups.remove("train")

        for grp in groups:
            all_samples.extend(_sample_sets_to_samples(self.references(grp)))
            all_samples.extend(_sample_sets_to_samples(self.probes(grp)))

        # Add znorm samples. Returning znorm samples for one group of dev or
        # eval is enough because they are duplicated.
        for grp in groups:
            all_samples.extend(_sample_sets_to_samples(self.zprobes(grp)))
            break

        # Add tnorm samples.
        all_samples.extend(_sample_sets_to_samples(self.treferences()))

        return all_samples

    def all_samples(self, groups=None):
        return self.samples(groups)

    @functools.lru_cache(maxsize=None)
    def _zprobes(self, protocol, group):
        sample_sets = self._sample_sets("train", "for_znorm")
        if not sample_sets:
            return sample_sets

        self._add_all_references(sample_sets, group)

        return sample_sets

    def zprobes(self, group="dev", proportion=1.0):
        sample_sets = self._zprobes(self.protocol, group)
        if not sample_sets:
            return sample_sets

        sample_sets = sample_sets[: int(len(sample_sets) * proportion)]
        return sample_sets

    @functools.lru_cache(maxsize=None)
    def _treferences(self, protocol):
        sample_sets = self._sample_sets("train", "for_tnorm")
        return sample_sets

    def treferences(self, proportion=1.0):
        sample_sets = self._treferences(self.protocol)
        if not sample_sets:
            return sample_sets

        sample_sets = sample_sets[: int(len(sample_sets) * proportion)]
        return sample_sets


class FileSampleLoader(BaseEstimator, TransformerMixin):
    """Loads file-based samples into :class:`~bob.pipelines.DelayedSample` objects.

    Given the :attr:`sample.path` attribute,``dataset_original_directory`` and an
    ``extension``, this transformer will load lazily the samples from the file.
    The ``data_loader`` is used to load the data.

    The resulting :class:`~bob.pipelines.DelayedSample` objects will call
    ``data_loader`` when their :attr:`data` is accessed.

    This transformer will not access the data files.

    Parameters
    ----------
    data_loader
        A callable to load the sample, given the full path to the file.
    dataset_original_directory
        Path of where the raw data files are stored. This will be prepended to the
        ``path`` attribute of the samples.
    extension
        File extension of the raw data files. This will be appended to the ``path``
        attribute of the samples.
    """

    def __init__(
        self,
        data_loader: Callable[[str], Any],
        dataset_original_directory: str = "",
        extension: str = "",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.data_loader = data_loader
        self.dataset_original_directory = dataset_original_directory
        self.extension = extension

    def transform(self, samples: list[Sample]) -> list[DelayedSample]:
        """Prepares the data into :class:`~bob.pipelines.DelayedSample` objects.

        Transforms :class:`~bob.pipelines.Sample` objects with a ``path`` attribute to
        :class:`~bob.pipelines.DelayedSample` with data ready to be loaded (lazily) by
        :attr:`data_loader`.

        When needed (access to the :class:`DelayedSample`\\ 's :attr:`data` attribute),
        :attr:`data_loader` will be called with the path (extended with
        :attr:`original_directory` and :attr:`extension`) as argument.

        Parameters
        ----------
        samples
            :class:`~bob.pipelines.Sample` objects with their ``path`` attribute
            containing a path to a file to load.
        """
        output = []
        for sample in samples:
            path = getattr(sample, "path")
            delayed_sample = DelayedSample(
                functools.partial(
                    self.data_loader,
                    os.path.join(
                        # we append ./ to path to make sure that the path is
                        # relative to the dataset_original_directory
                        self.dataset_original_directory,
                        f"./{path + self.extension}",
                    ),
                ),
                parent=sample,
            )
            output.append(delayed_sample)
        return output

    def _more_tags(self):
        return {"requires_fit": False}


class AnnotationsLoader(TransformerMixin, BaseEstimator):
    """Prepares annotations to be loaded from a path in ``delayed_attributes``.

    Metadata loader that loads samples' annotations using
    :py:func:`~bob.bio.base.utils.annotations.read_annotation_file`. This assumes that
    the annotation files follows the same folder structure and naming as the raw data
    files. Although, the base location and the extension can vary from those and is
    specified by :attr:`annotation_directory` and :attr:`annotation_extension`.

    Parameters
    ----------
    annotation_directory
        Path where the annotations are stored.

    annotation_extension : str
        Extension of the annotations.

    annotation_type : str
        Annotations type passed to
        :func:`~bob.bio.base.utils.annotations.read_annotation_file`.

    """

    def __init__(
        self,
        annotation_directory: Optional[str] = None,
        annotation_extension: str = ".json",
        annotation_type: str = "json",
    ):
        self.annotation_directory = annotation_directory
        self.annotation_extension = annotation_extension
        self.annotation_type = annotation_type

    def transform(self, X: list[DelayedSample]) -> list[DelayedSample]:
        """Edits the samples to lazily load annotations files.

        Parameters
        ----------
        X
            The samples to augment.
        """
        if self.annotation_directory is None:
            return None

        annotated_samples = []
        for x in X:
            # we use .key here because .path might not be unique for all
            # samples. Also, the ``bob bio annotate-samples`` command dictates
            # how annotations are stored.
            annotation_file = os.path.join(
                self.annotation_directory, x.key + self.annotation_extension
            )

            annotated_samples.append(
                DelayedSample.from_sample(
                    x,
                    delayed_attributes=dict(
                        annotations=lambda: read_annotation_file(
                            annotation_file, self.annotation_type
                        )
                    ),
                )
            )

        return annotated_samples

    def _more_tags(self):
        return {"requires_fit": False}



import os
import shutil
import tempfile

from click.testing import CliRunner

from bob.bio.base.annotator import Callable, FailSafe
from bob.bio.base.script.annotate import annotate, annotate_samples
from bob.bio.base.utils.annotations import read_annotation_file
from bob.io.base.testing_utils import assert_click_runner_result


def test_annotate():
    try:
        tmp_dir = tempfile.mkdtemp(prefix="bobtest_")
        runner = CliRunner()
        result = runner.invoke(
            annotate,
            args=("-d", "dummy", "-g", "dev", "-a", "dummy", "-o", tmp_dir),
        )
        assert_click_runner_result(result)

        # test if annotations exist
        for dirpath, _, filenames in os.walk(tmp_dir):
            for filename in filenames:
                path = os.path.join(dirpath, filename)
                annot = read_annotation_file(path, "json")
                assert annot["topleft"] == [0, 0]
                # size of atnt images
                assert annot["bottomright"] == [112, 92]
    finally:
        shutil.rmtree(tmp_dir)


def test_annotate_samples():
    try:
        tmp_dir = tempfile.mkdtemp(prefix="bobtest_")
        runner = CliRunner()
        result = runner.invoke(
            annotate_samples,
            args=("dummy_samples", "-a", "dummy", "-o", tmp_dir),
        )
        assert_click_runner_result(result)

        # test if annotations exist
        for dirpath, dirnames, filenames in os.walk(tmp_dir):
            for filename in filenames:
                path = os.path.join(dirpath, filename)
                annot = read_annotation_file(path, "json")
                assert annot["topleft"] == [0, 0]
                # size of atnt images
                assert annot["bottomright"] == [112, 92]
    finally:
        shutil.rmtree(tmp_dir)


def dummy_extra_key_annotator(data_batch, **kwargs):
    return [{"leye": 0, "reye": 0, "topleft": 0}]


def test_failsafe():
    annotator = FailSafe(
        [Callable(dummy_extra_key_annotator)], ["leye", "reye"]
    )
    annotations = annotator([1])
    assert all(key in annotations[0] for key in ["leye", "reye", "topleft"])

    annotator = FailSafe(
        [Callable(dummy_extra_key_annotator)], ["leye", "reye"], True
    )
    annotations = annotator([1])
    assert all(key in annotations[0] for key in ["leye", "reye"])
    assert all(key not in annotations[0] for key in ["topleft"])



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# @author: Tiago de Freitas Pereira <tiago.pereira@idiap.ch>

import copy
import os
import shutil
import tempfile
import uuid

import numpy as np
import pytest

from scipy.spatial.distance import cdist
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import FunctionTransformer

from bob.bio.base.algorithm import Distance
from bob.bio.base.pipelines import (
    CSVScoreWriter,
    PipelineScoreNorm,
    PipelineSimple,
    TNormScores,
    ZNormScores,
    checkpoint_pipeline_simple,
    dask_bio_pipeline,
)
from bob.bio.base.script.pipeline_score_norm import (
    pipeline_score_norm as pipeline_score_norm_cli,
)
from bob.pipelines import Sample, SampleSet
from tests.test_pipeline_simple import DummyDatabase, _make_transformer


def zt_norm_stubs(references, probes, t_references, z_probes):
    def _norm(scores, norm_base_scores, axis=1):
        mu = np.mean(norm_base_scores, axis=axis)

        old = True
        if old:
            std = np.std(norm_base_scores, axis=axis)
            if axis == 1:
                return ((scores.T - mu) / std).T
            else:
                return (scores - mu) / std

        if axis == 1:
            std = np.sqrt(
                np.sum(
                    (
                        norm_base_scores
                        - np.tile(
                            mu.reshape(norm_base_scores.shape[0], 1),
                            (1, norm_base_scores.shape[1]),
                        )
                    )
                    ** 2,
                    axis=1,
                )
                / (norm_base_scores.shape[1] - 1)
            )

            return (
                scores
                - np.tile(
                    mu.reshape(norm_base_scores.shape[0], 1),
                    (1, scores.shape[1]),
                )
            ) / np.tile(
                std.reshape(norm_base_scores.shape[0], 1), (1, scores.shape[1])
            )
        else:
            std = np.sqrt(
                np.sum(
                    (
                        norm_base_scores
                        - np.tile(
                            mu.reshape(1, norm_base_scores.shape[1]),
                            (norm_base_scores.shape[0], 1),
                        )
                    )
                    ** 2,
                    axis=0,
                )
                / (norm_base_scores.shape[0] - 1)
            )

            return (
                scores
                - np.tile(
                    mu.reshape(1, norm_base_scores.shape[1]),
                    (scores.shape[0], 1),
                )
            ) / np.tile(
                std.reshape(1, norm_base_scores.shape[1]), (scores.shape[0], 1)
            )

    n_reference = references.shape[0]
    n_probes = probes.shape[0]
    n_t_references = t_references.shape[0]
    n_z_probes = z_probes.shape[0]

    raw_scores = cdist(references, probes)

    z_scores = cdist(references, z_probes)
    # Computing the statistics of Z-Probes for each biometric reference
    # https://arxiv.org/pdf/1709.09868.pdf --> below eq (2) first eq
    z_normed_scores = _norm(raw_scores, z_scores, axis=1)
    assert z_normed_scores.shape == (n_reference, n_probes)

    t_scores = cdist(t_references, probes)
    # Computing the statistics of T-Models for each probe
    # https://arxiv.org/pdf/1709.09868.pdf --> below eq (2) second eq
    t_normed_scores = _norm(raw_scores, t_scores, axis=0)
    assert t_normed_scores.shape == (n_reference, n_probes)
    assert t_scores.shape == (n_t_references, n_probes)

    ZxT_scores = cdist(t_references, z_probes)
    assert ZxT_scores.shape == (n_t_references, n_z_probes)
    # Computing the statistics of T-Models for each z probe
    # https://arxiv.org/pdf/1709.09868.pdf --> below eq (2) third eq
    z_t_scores = _norm(t_scores, ZxT_scores, axis=1)
    assert z_t_scores.shape == (n_t_references, n_probes)

    # FINALLY DOING THE F*****G ZT-NORM
    zt_normed_scores = _norm(z_normed_scores, z_t_scores, axis=0)
    assert zt_normed_scores.shape == (n_reference, n_probes)

    s_normed_scores = (z_normed_scores + t_normed_scores) * 0.5
    assert s_normed_scores.shape == (n_reference, n_probes)

    return (
        raw_scores,
        z_normed_scores,
        t_normed_scores,
        zt_normed_scores,
        s_normed_scores,
    )


def _create_sample_sets(raw_data, offset, references=None):
    if references is None:
        return [
            SampleSet(
                [
                    Sample(
                        s,
                        template_id=str(i + offset),
                        key=str(uuid.uuid4()),
                    )
                ],
                key=str(i + offset),
                template_id=str(i + offset),
                subject_id=str(i + offset),
            )
            for i, s in enumerate(raw_data)
        ]
    else:
        return [
            SampleSet(
                [
                    Sample(
                        s,
                        template_id=str(i + offset),
                        key=str(uuid.uuid4()),
                    )
                ],
                key=str(i + offset),
                template_id=str(i + offset),
                subject_id=str(i + offset),
                references=references,
            )
            for i, s in enumerate(raw_data)
        ]


def _do_nothing_fn(x):
    return x


def _dump_scores_from_samples(scores, shape):
    # We have to transpose because the tests are BIOMETRIC_REFERENCES vs PROBES
    # and bob.bio.base is PROBES vs BIOMETRIC_REFERENCES
    return np.array([s.data for sset in scores for s in sset]).reshape(shape).T


def _make_pipeline_simple():
    transformer = make_pipeline(FunctionTransformer(func=_do_nothing_fn))
    biometric_algorithm = Distance("euclidean", factor=1)
    pipeline_simple = PipelineSimple(
        transformer, biometric_algorithm, score_writer=None
    )
    return pipeline_simple


def _run_norm_mechanics(with_dask, with_checkpoint=False, dir_name=None):
    ############
    # Prepating stubs
    ############

    n_references = 2
    n_probes = 3
    n_t_references = 4
    n_z_probes = 5
    dim = 5

    references = np.arange(n_references * dim).reshape(
        n_references, dim
    )  # two references (each row different identity)
    probes = (
        np.arange(n_probes * dim).reshape(n_probes, dim) * 10
    )  # three probes (each row different identity matching with references)

    t_references = np.arange(n_t_references * dim).reshape(
        n_t_references, dim
    )  # four T-REFERENCES (each row different identity)
    z_probes = (
        np.arange(n_z_probes * dim).reshape(n_z_probes, dim) * 10
    )  # five Z-PROBES (each row different identity matching with t references)

    (
        raw_scores_ref,
        z_normed_scores_ref,
        t_normed_scores_ref,
        zt_normed_scores_ref,
        s_normed_scores_ref,
    ) = zt_norm_stubs(references, probes, t_references, z_probes)

    ############
    # Preparing the samples
    ############

    # Creating enrollment samples
    biometric_reference_sample_sets = _create_sample_sets(references, offset=0)
    t_reference_sample_sets = _create_sample_sets(t_references, offset=300)

    # Fetching ids
    template_ids = [r.template_id for r in biometric_reference_sample_sets]
    # t_template_ids = [r.template_id for r in t_reference_sample_sets]
    # ids = template_ids + t_template_ids

    probe_sample_sets = _create_sample_sets(
        probes, offset=600, references=template_ids
    )
    z_probe_sample_sets = _create_sample_sets(
        z_probes, offset=900, references=template_ids
    )

    ############
    # TESTING REGULAR SCORING
    #############

    pipeline_simple = _make_pipeline_simple()
    if with_checkpoint:
        pipeline_simple = checkpoint_pipeline_simple(pipeline_simple, dir_name)

    if with_dask:
        pipeline_simple = dask_bio_pipeline(pipeline_simple)

    score_samples = pipeline_simple(
        [],
        biometric_reference_sample_sets,
        probe_sample_sets,
        score_all_vs_all=True,
    )

    if with_dask:
        score_samples = score_samples.compute(scheduler="single-threaded")

    raw_scores = _dump_scores_from_samples(
        score_samples, shape=(n_probes, n_references)
    )

    assert np.allclose(raw_scores, raw_scores_ref)

    ############
    # TESTING Z-NORM
    #############
    # """
    pipeline_simple = _make_pipeline_simple()
    z_norm_postprocessor = ZNormScores()

    z_pipeline_simple = PipelineScoreNorm(pipeline_simple, z_norm_postprocessor)

    if with_dask:
        z_pipeline_simple = dask_bio_pipeline(z_pipeline_simple)

    _, z_normed_score_samples = z_pipeline_simple(
        [],
        biometric_reference_sample_sets,
        copy.deepcopy(probe_sample_sets),
        z_probe_sample_sets,
    )

    z_normed_scores = _dump_scores_from_samples(
        z_normed_score_samples, shape=(n_probes, n_references)
    )

    np.testing.assert_allclose(z_normed_scores, z_normed_scores_ref)
    # """
    ############
    # TESTING T-NORM
    #############

    pipeline_simple = _make_pipeline_simple()
    t_norm_postprocessor = TNormScores()

    t_pipeline_simple = PipelineScoreNorm(pipeline_simple, t_norm_postprocessor)

    if with_dask:
        t_pipeline_simple = dask_bio_pipeline(t_pipeline_simple)

    _, t_normed_score_samples = t_pipeline_simple(
        [],
        biometric_reference_sample_sets,
        copy.deepcopy(probe_sample_sets),
        t_reference_sample_sets,
    )

    t_normed_scores = _dump_scores_from_samples(
        t_normed_score_samples, shape=(n_probes, n_references)
    )

    np.testing.assert_allclose(t_normed_scores, t_normed_scores_ref)


@pytest.mark.slow
def test_norm_mechanics():
    # No dask
    _run_norm_mechanics(with_dask=False)  # On memory

    # With checkpoing
    with tempfile.TemporaryDirectory() as dir_name:
        _run_norm_mechanics(
            with_dask=False, with_checkpoint=True, dir_name=dir_name
        )
        _run_norm_mechanics(
            with_dask=False, with_checkpoint=True, dir_name=dir_name
        )

    # With dask
    _run_norm_mechanics(with_dask=True)  # On memory
    with tempfile.TemporaryDirectory() as dir_name:
        _run_norm_mechanics(
            with_dask=True, with_checkpoint=True, dir_name=dir_name
        )
        _run_norm_mechanics(
            with_dask=True, with_checkpoint=True, dir_name=dir_name
        )


def test_znorm_on_memory():
    with tempfile.TemporaryDirectory() as dir_name:

        def run_pipeline(with_dask, score_writer=None):
            database = DummyDatabase(one_d=False)

            transformer = _make_transformer(dir_name)

            biometric_algorithm = Distance()

            pipeline_simple = PipelineSimple(
                transformer, biometric_algorithm, score_writer
            )

            z_norm_postprocessor = ZNormScores()

            z_pipeline_simple = PipelineScoreNorm(
                pipeline_simple, z_norm_postprocessor
            )

            # Checkpointing everything

            pipeline_simple = checkpoint_pipeline_simple(
                pipeline_simple, base_dir=dir_name
            )

            if with_dask:
                z_pipeline_simple = dask_bio_pipeline(
                    z_pipeline_simple, npartitions=2
                )

            (raw_scores, z_scores) = z_pipeline_simple(
                database.background_model_samples(),
                database.references(),
                database.probes(),
                database.zprobes(),
                score_all_vs_all=database.score_all_vs_all,
            )

            def _concatenate(pipeline, scores, path):
                writed_scores = pipeline.write_scores(scores)
                concatenated_scores = pipeline.post_process(writed_scores, path)
                return concatenated_scores

            if isinstance(score_writer, CSVScoreWriter):
                raw_scores = _concatenate(
                    z_pipeline_simple,
                    raw_scores,
                    os.path.join(dir_name, "scores-dev", "raw_scores"),
                )
                z_scores = _concatenate(
                    z_pipeline_simple,
                    z_scores,
                    os.path.join(dir_name, "scores-dev", "z_scores"),
                )
                """
                t_scores = _concatenate(
                    z_pipeline_simple,
                    t_scores,
                    os.path.join(dir_name, "scores-dev", "t_scores"),
                )

                zt_scores = _concatenate(
                    vanilla_biometrics_pipeline,
                    zt_scores,
                    os.path.join(dir_name, "scores-dev", "zt_scores"),
                )

                s_scores = _concatenate(
                    vanilla_biometrics_pipeline,
                    s_scores,
                    os.path.join(dir_name, "scores-dev", "s_scores"),
                )
                """

            if with_dask:
                raw_scores = raw_scores.compute(scheduler="single-threaded")
                z_scores = z_scores.compute(scheduler="single-threaded")
                # t_scores = t_scores.compute(scheduler="single-threaded")
                # zt_scores = zt_scores.compute(scheduler="single-threaded")
                # s_scores = s_scores.compute(scheduler="single-threaded")

            if isinstance(score_writer, CSVScoreWriter):
                assert (
                    len(
                        open(
                            os.path.join(dir_name, "scores-dev", "raw_scores"),
                            "r",
                        ).readlines()
                    )
                    == 101
                )
                assert (
                    len(
                        open(
                            os.path.join(dir_name, "scores-dev", "z_scores"),
                            "r",
                        ).readlines()
                    )
                    == 101
                )
                """
                assert (
                    len(
                        open(
                            os.path.join(dir_name, "scores-dev", "t_scores"), "r"
                        ).readlines()
                    )
                    == 101
                )
                assert (
                    len(
                        open(
                            os.path.join(dir_name, "scores-dev", "zt_scores"), "r"
                        ).readlines()
                    )
                    == 101
                )
                assert (
                    len(
                        open(
                            os.path.join(dir_name, "scores-dev", "s_scores"), "r"
                        ).readlines()
                    )
                    == 101
                )
                """

            else:
                assert len(raw_scores) == 10
                assert len(z_scores) == 10
                # assert len(t_scores) == 10
                # assert len(zt_scores) == 10
                # assert len(s_scores) == 10

        run_pipeline(False)
        run_pipeline(False)  # Testing checkpoint
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        run_pipeline(
            False, CSVScoreWriter(os.path.join(dir_name, "concatenated_scores"))
        )
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        # With DASK
        run_pipeline(True)
        run_pipeline(True)  # Testing checkpoint
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        run_pipeline(
            True, CSVScoreWriter(os.path.join(dir_name, "concatenated_scores"))
        )


def test_pipeline_score_norm_click_cli():
    from .test_pipeline_simple import _test_pipeline_click_cli

    for options in [
        ["--no-dask", "--memory"],
        ["--no-dask"],
        ["--memory"],
        [],
        ["--score-normalization-type", "tnorm"],
    ]:
        norm_output = "tnorm" if "tnorm" in options else "znorm"
        _test_pipeline_click_cli(
            pipeline_score_norm_cli,
            options,
            expected_outputs=[
                "results/scores-dev/raw_scores.csv",
                f"results/scores-dev/{norm_output}.csv",
            ],
        )



import numpy as np

from sklearn.decomposition import PCA
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import FunctionTransformer
from sklearn.utils import check_array

from bob.bio.base.pipelines import BioAlgorithm, PipelineSimple
from bob.pipelines import wrap

# Transformers #
pca = PCA(n_components=0.95)


# the images are in shape of Nx112x92, we want to flatten to Nx10304 them so we can train a PCA on them.
def flatten(images):
    images = check_array(images, allow_nd=True)
    new_shape = [images.shape[0], -1]
    return np.reshape(images, new_shape)


flatten_transformer = FunctionTransformer(flatten, validate=False)

# Chain the Transformers together
transformer = make_pipeline(flatten_transformer, pca)

# All transformers must be sample transformers
transformer = wrap(["sample"], transformer)


# Implementation of the BioAlgorithm #
# A generic implementation is available in:
# from bob.bio.base.algorithm import Distance
class EuclideanDistance(BioAlgorithm):
    def create_templates(self, list_of_feature_sets, enroll):
        # list_of_feature_sets is a list of lists of features with the shape of Nx?xD.
        # The second dimension can be variable in list_of_feature_sets depening on the
        # datbase protocol. Hence, you cannot call:
        # list_of_feature_sets = np.array(list_of_feature_sets)
        templates = []
        for feature_set in list_of_feature_sets:
            # template is a D dimensional vector
            template = np.mean(feature_set, axis=0)
            templates.append(template)
        # templates is an NxD dimensional array
        return templates

    def compare(self, enroll_templates, probe_templates):
        # enroll_templates is an NxD dimensional array
        # probe_templates is an MxD dimensional array
        # scores will be an NxM dimensional array
        scores = []
        for model in enroll_templates:
            scores.append([])
            for probe in probe_templates:
                similarity = 1 / np.linalg.norm(model - probe)
                scores[-1].append(similarity)
        scores = np.array(scores, dtype=float)
        return scores


bio_algorithm = EuclideanDistance()

# Creation of the pipeline #
# `pipeline` will be used by the `bob bio pipeline simple` command
pipeline = PipelineSimple(transformer, bio_algorithm)

# you can also specify the other options in this file:
database = "atnt"
output = "results"



# isort: skip_file
from .pipelines import PipelineSimple  # noqa: F401
from .score_writers import FourColumnsScoreWriter, CSVScoreWriter
from .wrappers import (  # noqa: F401
    BioAlgCheckpointWrapper,
    BioAlgDaskWrapper,
    checkpoint_pipeline_simple,
    dask_bio_pipeline,
    is_biopipeline_checkpointed,
    get_bio_alg_tags,
)

from .abstract_classes import BioAlgorithm, Database, ScoreWriter

from .score_post_processor import (  # noqa: F401
    PipelineScoreNorm,
    ZNormScores,
    TNormScores,
    CategoricalCalibration,
    WeibullCalibration,
    LLRCalibration,
    GammaCalibration,
    BetaCalibration,
)

from .entry_points import (  # noqa: F401
    execute_pipeline_simple,
    execute_pipeline_score_norm,
    execute_pipeline_train,
)


# gets sphinx autodoc done right - don't remove it
def __appropriate__(*args):
    """Says object was actually declared here, and not in the import module.
    Fixing sphinx warnings of not being able to find classes, when path is
    shortened.

    Parameters
    ----------
    *args
        An iterable of objects to modify

    Resolves `Sphinx referencing issues
    <https://github.com/sphinx-doc/sphinx/issues/3048>`
    """

    for obj in args:
        obj.__module__ = __name__


__appropriate__(
    PipelineSimple,
    FourColumnsScoreWriter,
    CSVScoreWriter,
    BioAlgCheckpointWrapper,
    BioAlgDaskWrapper,
    BioAlgorithm,
    Database,
    ScoreWriter,
    PipelineScoreNorm,
    ZNormScores,
    TNormScores,
    CategoricalCalibration,
    WeibullCalibration,
    LLRCalibration,
    GammaCalibration,
    BetaCalibration,
)

__all__ = [_ for _ in dir() if not _.startswith("_")]



import logging
import pickle

from bob.bio.base.pipelines import BioAlgorithm
from bob.learn.em import ISVMachine

logger = logging.getLogger(__name__)


class ISV(ISVMachine, BioAlgorithm):
    """ISV transformer and bioalgorithm to be used in pipelines"""

    def transform(self, X):
        """Passthrough"""
        return X

    def create_templates(self, list_of_feature_sets, enroll):
        if enroll:
            return [
                self.enroll(feature_set) for feature_set in list_of_feature_sets
            ]
        else:
            # TODO: We should compute these parts of self.score:
            # x = self.estimate_x(data)
            # Ux = self._U @ x
            # here to make scoring faster
            return list_of_feature_sets

    def compare(self, enroll_templates, probe_templates):
        # TODO: The underlying score method actually supports batched scoring
        return [
            [self.score(enroll, probe) for probe in probe_templates]
            for enroll in enroll_templates
        ]

    @classmethod
    def custom_enrolled_save_fn(cls, data, path):
        pickle.dump(data, open(path, "wb"))

    @classmethod
    def custom_enrolled_load_fn(cls, path):
        return pickle.load(open(path, "rb"))

    def _more_tags(self):
        return {
            "bob_fit_supports_dask_bag": True,
            "bob_fit_extra_input": [("y", "subject_id_int")],
            "bob_enrolled_save_fn": self.custom_enrolled_save_fn,
            "bob_enrolled_load_fn": self.custom_enrolled_load_fn,
            "bob_checkpoint_features": False,
        }



"""The main entry for bob.bio (click-based) scripts.
"""
import click
import pkg_resources

from clapper.click import AliasedGroup
from click_plugins import with_plugins


@with_plugins(pkg_resources.iter_entry_points("bob.bio.cli"))
@click.group(cls=AliasedGroup)
def bio():
    """Biometric recognition commands."""
    pass



import numpy

import bob.bio.base

from bob.bio.base.database import BioFile
from bob.bio.base.extractor import Extractor

_data = [0.0, 1.0, 2.0, 3.0, 4.0]


class DummyExtractor(Extractor):
    def __init__(self, **kwargs):
        Extractor.__init__(self, requires_training=True)
        self.model = False

    def train(self, train_data, extractor_file):
        assert isinstance(train_data, list)
        bob.bio.base.save(_data, extractor_file)

    def load(self, extractor_file):
        data = bob.bio.base.load(extractor_file)
        assert (_data == data).all()
        self.model = True

    def __call__(self, data):
        """Does nothing, simply converts the data type of the data, ignoring any annotation."""
        assert self.model
        return data.astype(numpy.float).flatten()


extractor = DummyExtractor()


class DummyExtractorMetadata(DummyExtractor):
    def __call__(self, data, metadata=None):
        """Does nothing, simply converts the data type of the data, ignoring any annotation."""
        assert isinstance(metadata, BioFile)
        return super(DummyExtractorMetadata, self).__call__(data)


extractor_metadata = DummyExtractorMetadata()



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# @author: Manuel Guenther <Manuel.Guenther@idiap.ch>
# @date: Thu May 24 10:41:42 CEST 2012

import numpy as np
import pkg_resources

import bob.bio.base
import bob.io.base

from . import utils


def test_distance_algorithm():
    # test the two registered distance functions

    # euclidean distance
    euclidean = bob.bio.base.algorithm.Distance(distance_function="euclidean")
    assert isinstance(euclidean, bob.bio.base.algorithm.Distance)

    # test distance computation
    f1 = np.ones((20, 10), np.float64)
    f2 = np.ones((20, 10), np.float64) * 2.0

    models = euclidean.create_templates([[f1, f1]], enroll=True)
    probes = euclidean.create_templates([[f2, f2]], enroll=False)
    score = euclidean.compare(models, probes)[0, 0]
    np.testing.assert_almost_equal(score, -np.sqrt(10))

    # test cosine distance
    cosine = bob.bio.base.algorithm.Distance(distance_function="cosine")
    models = cosine.create_templates([[f1, f1]], enroll=True)
    probes = cosine.create_templates([[f2, f2]], enroll=False)
    score = cosine.compare(models, probes)[0, 0]
    np.testing.assert_almost_equal(score, 0)

    # compare model with probe
    enroll = utils.random_training_set(5, 5, 0.0, 255.0, seed=21)
    probe = bob.io.base.load(
        pkg_resources.resource_filename(__name__, "data/lda_projected.hdf5")
    )
    models = cosine.create_templates([enroll], enroll=True)
    probes = cosine.create_templates([probe], enroll=False)
    score = cosine.compare(models, probes)[0, 0]

    np.testing.assert_almost_equal(score, -0.1873371)



import functools
import logging
import os

from typing import Any, Callable, Optional

import dask
import h5py
import numpy as np

from sklearn.pipeline import Pipeline

import bob.pipelines

from bob.bio.base.pipelines import PipelineSimple
from bob.pipelines import (
    CheckpointWrapper,
    DelayedSample,
    Sample,
    is_instance_nested,
)
from bob.pipelines.wrappers import BaseWrapper, _frmt, get_bob_tags

from .abstract_classes import BioAlgorithm

logger = logging.getLogger(__name__)


def default_save(data: np.ndarray, path: str):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with h5py.File(path, "w") as f:
        f["data"] = data


def default_load(path: str) -> np.ndarray:
    with h5py.File(path, "r") as f:
        return f["data"][()]


def get_bio_alg_tags(estimator=None, force_tags=None):
    bob_tags = get_bob_tags(estimator=estimator, force_tags=force_tags)
    default_tags = {
        "bob_enrolled_extension": ".h5",
        "bob_enrolled_save_fn": default_save,
        "bob_enrolled_load_fn": default_load,
    }
    force_tags = force_tags or {}
    estimator_tags = estimator._get_tags() if estimator is not None else {}
    return {**bob_tags, **default_tags, **estimator_tags, **force_tags}


class BioAlgorithmBaseWrapper(BioAlgorithm, BaseWrapper):
    def create_templates(self, feature_sets, enroll):
        return self.biometric_algorithm.create_templates(feature_sets, enroll)

    def compare(self, enroll_templates, probe_templates):
        return self.biometric_algorithm.compare(
            enroll_templates, probe_templates
        )


class BioAlgCheckpointWrapper(BioAlgorithmBaseWrapper):
    """Wrapper used to checkpoint enrolled and Scoring samples.

    Parameters
    ----------
    biometric_algorithm
        An implemented :any:`BioAlgorithm`

    base_dir
        Path to store biometric references and scores

    extension
        Default extension of the enrolled references files.
        If None, will use the ``bob_checkpoint_extension`` tag in the estimator, or
        default to ``.h5``.

    save_func
        Pointer to a customized function that saves an enrolled reference to the disk.
        If None, will use the ``bob_enrolled_save_fn`` tag in the estimator, or default
        to h5py.

    load_func
        Pointer to a customized function that loads an enrolled reference from disk.
        If None, will use the ``bob_enrolled_load_fn`` tag in the estimator, or default
        to h5py.

    group
        group of the data, used to save different group's checkpoints in different dirs.

    force
        If True, will recompute scores and biometric references no matter if a file
        exists

    hash_fn
        Pointer to a hash function. This hash function maps
        `sample.key` to a hash code and this hash code corresponds a relative directory
        where a single `sample` will be checkpointed.
        This is useful when is desirable file directories with less than a certain
        number of files.

    Examples
    --------

    >>> from bob.bio.base.algorithm import Distance
    >>> from bob.bio.base.pipelines import BioAlgCheckpointWrapper
    >>> biometric_algorithm = BioAlgCheckpointWrapper(Distance(), base_dir="./")
    >>> biometric_algorithm.create_templates(samples, enroll=True) # doctest: +SKIP

    """

    def __init__(
        self,
        biometric_algorithm: BioAlgorithm,
        base_dir: str,
        extension: Optional[str] = None,
        save_func: Optional[Callable[[str, Any], None]] = None,
        load_func: Optional[Callable[[str], Any]] = None,
        group: Optional[str] = None,
        force: bool = False,
        hash_fn: Optional[Callable[[str], str]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.base_dir = base_dir
        self.set_score_references_path(group)
        self.group = group
        self.biometric_algorithm = biometric_algorithm
        self.force = force
        self.hash_fn = hash_fn
        bob_tags = get_bio_alg_tags(self.biometric_algorithm)
        self.extension = extension or bob_tags["bob_enrolled_extension"]
        self.save_func = save_func or bob_tags["bob_enrolled_save_fn"]
        self.load_func = load_func or bob_tags["bob_enrolled_load_fn"]

    def set_score_references_path(self, group):
        if group is None:
            self.biometric_reference_dir = os.path.join(
                self.base_dir, "biometric_references"
            )
        else:
            self.biometric_reference_dir = os.path.join(
                self.base_dir, group, "biometric_references"
            )

    def write_biometric_reference(self, sample, path):
        data = sample.data
        if data is None:
            raise RuntimeError("Cannot checkpoint template of None")
        return self.save_func(sample.data, path)

    def _enroll_sample_set(self, sampleset):
        """
        Enroll a sample set with checkpointing
        """

        # If sampleset has a key use it, otherwise use the first sample's key
        model_key = getattr(sampleset, "key", sampleset.samples[0].key)

        # Amending `models` directory
        hash_dir_name = (
            self.hash_fn(str(model_key)) if self.hash_fn is not None else ""
        )

        path = os.path.join(
            self.biometric_reference_dir,
            hash_dir_name,
            str(model_key) + self.extension,
        )

        if self.force or not os.path.exists(path):
            enrolled_sample = (
                self.biometric_algorithm.create_templates_from_samplesets(
                    [sampleset], enroll=True
                )[0]
            )

            # saving the new sample
            os.makedirs(os.path.dirname(path), exist_ok=True)
            self.write_biometric_reference(enrolled_sample, path)

        # This seems inefficient, but it's crucial for large datasets
        delayed_enrolled_sample = DelayedSample(
            functools.partial(self.load_func, path), parent=sampleset
        )

        return delayed_enrolled_sample

    def create_templates_from_samplesets(self, list_of_samplesets, enroll):
        logger.debug(
            f"{_frmt(self, attr='biometric_algorithm')}.create_templates_from_samplesets(... enroll={enroll})"
        )
        if not enroll:
            return self.biometric_algorithm.create_templates_from_samplesets(
                list_of_samplesets, enroll
            )
        retval = []
        for sampleset in list_of_samplesets:
            # if it exists, load it!
            sample = self._enroll_sample_set(sampleset)
            retval.append(sample)
        return retval


class BioAlgDaskWrapper(BioAlgorithmBaseWrapper):
    """
    Wrap :any:`bob.bio.base.pipelines.BioAlgorithm` to work with DASK
    """

    def __init__(self, biometric_algorithm: BioAlgorithm, **kwargs):
        self.biometric_algorithm = biometric_algorithm

    def create_templates_from_samplesets(self, list_of_samplesets, enroll):
        logger.debug(
            f"{_frmt(self, attr='biometric_algorithm')}.create_templates_from_samplesets(... enroll={enroll})"
        )
        templates = list_of_samplesets.map_partitions(
            self.biometric_algorithm.create_templates_from_samplesets,
            enroll=enroll,
        )
        return templates

    def score_sample_templates(
        self, probe_samples, enroll_samples, score_all_vs_all
    ):
        logger.debug(
            f"{_frmt(self, attr='biometric_algorithm')}.score_sample_templates(... score_all_vs_all={score_all_vs_all})"
        )
        # load the templates into memory because they could be delayed samples
        enroll_samples = enroll_samples.map_partitions(
            _delayed_samples_to_samples
        )
        probe_samples = probe_samples.map_partitions(
            _delayed_samples_to_samples
        )

        all_references = dask.delayed(list)(enroll_samples)
        scores = probe_samples.map_partitions(
            self.biometric_algorithm.score_sample_templates,
            all_references,
            score_all_vs_all=score_all_vs_all,
        )
        return scores


def _delayed_samples_to_samples(delayed_samples):
    return [Sample(sample.data, parent=sample) for sample in delayed_samples]


def dask_bio_pipeline(
    pipeline: PipelineSimple,
    npartitions: Optional[int] = None,
    partition_size: Optional[int] = None,
):
    """
    Given a :any:`PipelineSimple`, wraps their :attr:`transformer` and
    :attr:`biometric_algorithm` to be executed with dask.

    Parameters
    ----------

    pipeline
       pipeline to be dasked

    npartitions
       Number of partitions for the initial `dask.bag`

    partition_size
       Size of the partition for the initial `dask.bag`
    """
    dask_wrapper_kw = {}
    if partition_size is None:
        dask_wrapper_kw["npartitions"] = npartitions
    else:
        dask_wrapper_kw["partition_size"] = partition_size

    pipeline.transformer = bob.pipelines.wrap(
        ["dask"], pipeline.transformer, **dask_wrapper_kw
    )
    pipeline.biometric_algorithm = BioAlgDaskWrapper(
        pipeline.biometric_algorithm
    )

    def _write_scores(scores):
        return scores.map_partitions(pipeline.write_scores_on_dask)

    pipeline.write_scores_on_dask = pipeline.write_scores
    pipeline.write_scores = _write_scores

    if hasattr(pipeline, "post_processor"):
        # cannot use bob.pipelines.wrap here because the input is already a dask bag.
        pipeline.post_processor = bob.pipelines.DaskWrapper(
            pipeline.post_processor
        )

    return pipeline


def checkpoint_pipeline_simple(
    pipeline: PipelineSimple,
    base_dir: str,
    biometric_algorithm_dir: Optional[str] = None,
    hash_fn: Optional[Callable[[str], str]] = None,
    force: bool = False,
):
    """
    Given a :any:`PipelineSimple`, wraps their :attr:`transformer` and
    :attr:`biometric_algorithm` to be checkpointed.

    If an estimator of the pipeline is already checkpointed, it will not be
    wrapped again.

    Parameters
    ----------

    pipeline
       pipeline to be checkpointed

    base_dir
       Path to store transformed input data and possibly biometric references and scores

    biometric_algorithm_dir
       If set, it will checkpoint the biometric references and scores to this path.
       If not, `base_dir` will be used.
       This is useful when it's suitable to have the transformed data path, and biometric references and scores
       in different paths.

    hash_fn
       Pointer to a hash function. This hash function will map
       `sample.key` to a hash code and this hash code will be the
       relative directory where a single `sample` will be checkpointed.
       This is useful when is desireable file directories with more than
       a certain number of files.
    force
       Overwrite existing checkpoint files.
    """

    bio_ref_scores_dir = (
        base_dir if biometric_algorithm_dir is None else biometric_algorithm_dir
    )

    if isinstance(pipeline.transformer, Pipeline):
        for step, (name, transformer) in enumerate(pipeline.transformer.steps):
            if not is_instance_nested(
                transformer, "estimator", CheckpointWrapper
            ):
                pipeline.transformer.steps[step] = (
                    name,
                    bob.pipelines.wrap(
                        ["checkpoint"],
                        transformer,
                        features_dir=os.path.join(base_dir, name),
                        model_path=os.path.join(base_dir, name),
                        hash_fn=hash_fn,
                        force=force,
                    ),
                )
    else:  # The pipeline.transformer is a lone transformer
        if not is_instance_nested(
            pipeline.transformer, "estimator", CheckpointWrapper
        ):
            pipeline.transformer = bob.pipelines.wrap(
                ["checkpoint"],
                pipeline.transformer,
                features_dir=base_dir,
                model_path=base_dir,
                hash_fn=hash_fn,
                force=force,
            )

    pipeline.biometric_algorithm = BioAlgCheckpointWrapper(
        pipeline.biometric_algorithm,
        base_dir=bio_ref_scores_dir,
        hash_fn=hash_fn,
        force=force,
    )

    return pipeline


def is_biopipeline_checkpointed(pipeline: PipelineSimple) -> bool:
    """
    Check if :any:`PipelineSimple` is checkpointed

    Parameters
    ----------

    pipeline
       pipeline to check if checkpointed by a :any:`BioAlgCheckpointWrapper`.

    """

    # We have to check if biometric_algorithm is checkpointed
    return is_instance_nested(
        pipeline, "biometric_algorithm", BioAlgCheckpointWrapper
    )



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Tiago de Freitas Pereira <tiago.pereira@idiap.ch>


"""Executes biometric pipeline"""

import logging

import click

from clapper.click import ConfigCommand, ResourceOption, verbosity_option

from bob.pipelines.distributed import VALID_DASK_CLIENT_STRINGS

logger = logging.getLogger(__name__)


EPILOG = """\b

Command line examples\n
-----------------------

$ bob bio pipeline simple -vv DATABASE PIPELINE

See the help of the CONFIG argument on top of this help message
for a list of available configurations.

It is possible to provide database and pipeline through a configuration file.
Generate an example configuration file with:

$ bob bio pipeline simple --dump-config my_experiment.py

and execute it with:

$ bob bio pipeline simple -vv my_experiment.py

my_experiment.py must contain the following elements:

   >>> transformer = ... # A scikit-learn pipeline wrapped with bob.pipelines' SampleWrapper\n
   >>> algorithm   = ... # `An BioAlgorithm`\n
   >>> pipeline = PipelineSimple(transformer,algorithm)\n
   >>> database = .... # Biometric Database (class that implements the methods: `background_model_samples`, `references` and `probes`)"
\b"""


@click.command(
    name="simple",
    entry_point_group="bob.bio.config",
    cls=ConfigCommand,
    epilog=EPILOG,
)
@click.option(
    "--pipeline",
    "-p",
    required=True,
    entry_point_group="bob.bio.pipeline",
    help="The simplest pipeline possible composed of a scikit-learn Pipeline and a BioAlgorithm",
    cls=ResourceOption,
)
@click.option(
    "--database",
    "-d",
    entry_point_group="bob.bio.database",
    required=True,
    help="Biometric Database connector (class that implements the methods: `background_model_samples`, `references` and `probes`)",
    cls=ResourceOption,
)
@click.option(
    "--dask-client",
    "-l",
    entry_point_group="dask.client",
    string_exceptions=VALID_DASK_CLIENT_STRINGS,
    default="single-threaded",
    help="Dask client for the execution of the pipeline.",
    cls=ResourceOption,
)
@click.option(
    "--group",
    "-g",
    "groups",
    type=click.Choice(["dev", "eval"]),
    multiple=True,
    default=("dev",),
    help="If given, this value will limit the experiments belonging to a particular protocolar group",
    cls=ResourceOption,
)
@click.option(
    "--output",
    "-o",
    show_default=True,
    default="results",
    help="Name of output directory where output scores will be saved.",
    cls=ResourceOption,
)
@click.option(
    "--write-metadata-scores/--write-column-scores",
    "-meta/-nmeta",
    default=True,
    help="If set, all the scores will be written with all their metadata using the `CSVScoreWriter`",
    cls=ResourceOption,
)
@click.option(
    "--memory",
    "-m",
    is_flag=True,
    help="If set, it will run the experiment keeping all objects on memory with nothing checkpointed. If not set, checkpoints will be saved in `--output`.",
    cls=ResourceOption,
)
@click.option(
    "--checkpoint-dir",
    "-c",
    show_default=True,
    default=None,
    help="Name of output directory where the checkpoints will be saved. In case --checkpoint is set, checkpoints will be saved in this directory.",
    cls=ResourceOption,
)
@click.option(
    "--dask-partition-size",
    "-s",
    help="If using Dask, this option defines the max size of each dask.bag.partition. "
    "Use this option if the current heuristic that sets this value doesn't suit your experiment. "
    "(https://docs.dask.org/en/latest/bag-api.html?highlight=partition_size#dask.bag.from_sequence).",
    default=None,
    type=click.INT,
    cls=ResourceOption,
)
@click.option(
    "--dask-n-partitions",
    "-n",
    help="If using Dask, this option defines a fixed number of dask.bag.partition for "
    "each set of data. Use this option if the current heuristic that sets this value "
    "doesn't suit your experiment."
    "(https://docs.dask.org/en/latest/bag-api.html?highlight=partition_size#dask.bag.from_sequence).",
    default=None,
    type=click.INT,
    cls=ResourceOption,
)
@click.option(
    "--dask-n-workers",
    "-w",
    help="If using Dask, this option defines the number of workers to start your experiment. "
    "Dask automatically scales up/down the number of workers due to the current load of tasks to be solved. "
    "Use this option if the current amount of workers set to start an experiment doesn't suit you.",
    default=None,
    type=click.INT,
    cls=ResourceOption,
)
@click.option(
    "--force",
    "-f",
    is_flag=True,
    help="If set, it will force generate all the checkpoints of an experiment. This option doesn't work if `--memory` is set",
    cls=ResourceOption,
)
@click.option(
    "--no-dask",
    is_flag=True,
    help="If set, it will not use Dask to run the experiment.",
    cls=ResourceOption,
)
@verbosity_option(cls=ResourceOption, logger=logger)
def pipeline_simple(
    pipeline,
    database,
    dask_client,
    groups,
    output,
    write_metadata_scores,
    memory,
    checkpoint_dir,
    dask_partition_size,
    dask_n_workers,
    dask_n_partitions,
    force,
    no_dask,
    **kwargs,
):
    """Runs the simplest biometrics pipeline.

    Such pipeline consists into two major components.
    The first component consists of a scikit-learn `Pipeline`,
    where a sequence of transformations of the input data
    is defined.
    The second component is a `BioAlgorithm` that defines the primitives
    `enroll` and `score`

    With those two components any Biometric Experiment can be done.
    A Biometric experiment consists of three sub-pipelines and
    they are defined below:

    Sub-pipeline 1:\n
    ---------------

    Training background model.
    Some biometric algorithms demands the training of background model, for instance a neural network.

    \b
    This pipeline runs: `Pipeline.fit(DATA_FOR_FIT)`



    \b

    Sub-pipeline 2:\n
    ---------------

    Creation of biometric references: This is a standard step in a biometric pipelines.
    Given a set of samples of one identity, create a biometric reference (a.k.a template) for sub identity.


    \b
    raw_data --> preprocessing >> feature extraction >> enroll(background_model) --> biometric_reference

    This pipeline runs: `BioAlgorithm.enroll(Pipeline.transform(DATA_ENROLL))` >> biometric_references


    Sub-pipeline 3:\n
    ---------------

    Probing: This is another standard step in biometric pipelines.
    Given one sample and one biometric reference, computes a score.
    Such score has different meanings depending on the scoring method your biometric algorithm uses.
    It's out of scope to explain in a help message to explain what scoring is for different biometric algorithms.

    This pipeline runs: `BioAlgorithm.score(Pipeline.transform(DATA_SCORE, biometric_references))` >> biometric_references

    .. Note::
        Refrain from calling this function directly from a script. Prefer
        :py:func:`bob.bio.base.pipelines.execute_pipeline_simple`
        instead.


    Using Dask
    ----------

    This pipeline is intended to work with Dask to split the load of work between
    processes on a machine or workers on a distributed grid system. By default, the
    local machine is used in single-threaded mode. However, by specifying the
    `--dask-client` option, you specify a Dask Client.

    When using multiple workers, a few things have to be considered:
    - The number of partitions in the data.
    - The number of workers to process the data.

    Ideally, (and this is the default behavior) you want to split all the data between
    many available workers, and all the workers work at the same time on all the data.
    But the number of workers may be limited, or one partition of data may be filling
    the memory of one worker. Moreover, having many small tasks (by splitting the data
    into many partitions) is not recommended as the scheduler will then spend more time
    organizing and communicating with the workers.

    To solve speed or memory issues, options are available to split the data
    differently (`--dask-n-partitions` or `--dask-partition-size`). If you encounter
    memory issues on a worker, try augmenting the number of partitions, and if your
    scheduler is not keeping up, try reducing that number.
    """
    from bob.bio.base.pipelines import execute_pipeline_simple

    if no_dask:
        dask_client = None

    checkpoint = not memory

    logger.debug("Executing PipelineSimple with:")
    logger.debug(f"pipeline: {pipeline}")
    logger.debug(f"  transformer: {pipeline.transformer}")
    logger.debug(f"  biometric_algorithm: {pipeline.biometric_algorithm}")
    logger.debug(f"database: {database}")

    execute_pipeline_simple(
        pipeline=pipeline,
        database=database,
        dask_client=dask_client,
        groups=groups,
        output=output,
        write_metadata_scores=write_metadata_scores,
        checkpoint=checkpoint,
        dask_partition_size=dask_partition_size,
        dask_n_partitions=dask_n_partitions,
        dask_n_workers=dask_n_workers,
        checkpoint_dir=checkpoint_dir,
        force=force,
    )

    logger.info("Experiment finished !")



from sklearn.base import BaseEstimator

from bob.bio.base.algorithm import Distance
from bob.bio.base.pipelines import PipelineSimple
from bob.pipelines import wrap


class DummyTransformer(BaseEstimator):
    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return [X[0].mean(axis=1)]


transformer = wrap(["sample"], DummyTransformer())
biometric_algorithm = Distance()

pipeline = PipelineSimple(transformer, biometric_algorithm)



"""Sorts score files based on their score value
"""
import logging

import click
import numpy

from clapper.click import log_parameters, verbosity_option

from bob.bio.base.score.load import dump_score, load_score

logger = logging.getLogger(__name__)


@click.command(
    epilog="""\b
Examples:

  $ bob bio sort -vvv /path/to/scores
"""
)
@click.argument(
    "score_paths",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, writable=True),
    nargs=-1,
)
@verbosity_option(logger=logger)
def sort(score_paths, **kwargs):
    """Sorts score files based on their score values

    The conversion happens in-place; backup your scores before using this script
    """
    log_parameters(logger)

    for path in score_paths:
        logger.info("Sorting: %s", path)
        scores = load_score(path)
        scores = scores[numpy.argsort(scores["score"])]
        dump_score(path, scores)



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Tiago de Freitas Pereira <tiago.pereira@idiap.ch>


"""Executes biometric pipeline"""

import functools
import logging

from typing import List

import click
import dask.distributed

from clapper.click import ResourceOption, verbosity_option
from tabulate import tabulate

import bob.io.base

from bob.bio.base.pipelines import PipelineSimple, dask_bio_pipeline
from bob.pipelines import DelayedSample, Sample, SampleSet

logger = logging.getLogger(__name__)

EPILOG = """\n



 Command line examples\n
 -----------------------

    >>> bob bio compare-samples ./imgs/1.png ./imgs/2.png -p inception_resnetv2_msceleb \n
    \n
    \n

    All vs All comparison \n
    -------------------  ------------------- \n
    ./imgs/1.png         ./imgs/2.png         \n
    0.0                  -0.5430189337666903  \n
    -0.5430189337666903  0.0                  \n
    -------------------  -------------------  \n

"""


@click.command(epilog=EPILOG)
@click.argument("samples", nargs=-1)
@click.option(
    "--pipeline",
    "-p",
    required=True,
    cls=ResourceOption,
    entry_point_group="bob.bio.pipeline",
    help="Vanilla biometrics pipeline composed of a scikit-learn Pipeline and a BioAlgorithm",
)
@click.option(
    "--dask-client",
    "-l",
    required=False,
    cls=ResourceOption,
    entry_point_group="dask.client",
    help="Dask client for the execution of the pipeline.",
)
@verbosity_option(logger=logger)
def compare_samples(
    samples: List[Sample],
    pipeline: PipelineSimple,
    dask_client: dask.distributed.Client,
    verbose: int,
):
    """Compare several samples in a All vs All fashion."""
    if len(samples) == 1:
        raise ValueError(
            "It's necessary to have at least two samples for the comparison"
        )

    sample_sets = [
        SampleSet(
            [DelayedSample(functools.partial(bob.io.base.load, s), key=str(s))],
            key=str(s),
            biometric_id=str(i),
        )
        for i, s in enumerate(samples)
    ]
    if dask_client is not None:
        pipeline = dask_bio_pipeline(pipeline)

    table = [[s for s in samples]]
    enroll_templates = pipeline.enroll_templates(sample_sets)
    probe_templates = pipeline.probe_templates(sample_sets)
    scores = pipeline.compute_scores(
        probe_templates, enroll_templates, score_all_vs_all=True
    )
    if dask_client is not None:
        scores = scores.compute(scheduler=dask_client)
    for sset in scores:
        table.append([str(s.data) for s in sset])

    print("All vs All comparison")
    print(tabulate(table))

    if dask_client is not None:
        dask_client.shutdown()



"""A script to help annotate databases.
"""
import functools
import json
import logging

import click

from clapper.click import (
    ConfigCommand,
    ResourceOption,
    log_parameters,
    verbosity_option,
)

from bob.pipelines import DelayedSample, ToDaskBag, wrap

logger = logging.getLogger(__name__)


def save_json(data, path):
    """
    Saves a dictionary ``data`` in a json file at ``path``.
    """
    with open(path, "w") as f:
        json.dump(data, f)


def load_json(path):
    """
    Returns a dictionary from a json file at ``path``.
    """
    with open(path, "r") as f:
        return json.load(f)


def annotate_common_options(func):
    @click.option(
        "--annotator",
        "-a",
        required=True,
        cls=ResourceOption,
        entry_point_group="bob.bio.annotator",
        help="An annotator (instance of class inheriting from "
        "bob.bio.base.Annotator) or an annotator resource name.",
    )
    @click.option(
        "--output-dir",
        "-o",
        required=True,
        cls=ResourceOption,
        help="The directory to save the annotations.",
    )
    @click.option(
        "--dask-client",
        "-l",
        "dask_client",
        entry_point_group="dask.client",
        help="Dask client for the execution of the pipeline. If not specified, "
        "uses a single threaded, local Dask Client.",
        cls=ResourceOption,
    )
    @functools.wraps(func)
    def wrapper(*args, **kwds):
        return func(*args, **kwds)

    return wrapper


@click.command(
    entry_point_group="bob.bio.config",
    cls=ConfigCommand,
    epilog="""\b
Examples:

  $ bob bio annotate -vvv -d <database> -a <annotator> -o /tmp/annotations
""",
)
@click.option(
    "--database",
    "-d",
    required=True,
    cls=ResourceOption,
    entry_point_group="bob.bio.database",
    help="Biometric Database (class that implements the methods: "
    "`background_model_samples`, `references` and `probes`).",
)
@click.option(
    "--groups",
    "-g",
    multiple=True,
    help="Biometric Database group that will be annotated. Can be added multiple"
    "times for different groups. [Default: All groups]",
)
@annotate_common_options
@verbosity_option(logger=logger, expose_value=False)
def annotate(database, groups, annotator, output_dir, dask_client, **kwargs):
    """Annotates a database.

    The annotations are written in text file (json) format which can be read
    back using :any:`read_annotation_file` (annotation_type='json')
    """
    log_parameters(logger)

    # Allows passing of Sample objects as parameters
    annotator = wrap(["sample"], annotator, output_attribute="annotations")

    # Will save the annotations in the `data` fields to a json file
    annotator = wrap(
        ["checkpoint"],
        annotator,
        features_dir=output_dir,
        extension=".json",
        save_func=save_json,
        load_func=load_json,
        sample_attribute="annotations",
    )

    # Allows reception of Dask Bags
    annotator = wrap(["dask"], annotator)

    # Transformer that splits the samples into several Dask Bags
    to_dask_bags = ToDaskBag(npartitions=50)

    logger.debug("Retrieving samples from database.")
    samples = database.all_samples(groups)

    # Sets the scheduler to local if no dask_client is specified
    if dask_client is not None:
        scheduler = dask_client
    else:
        scheduler = "single-threaded"

    # Splits the samples list into bags
    dask_bags = to_dask_bags.transform(samples)

    logger.info(f"Saving annotations in {output_dir}.")
    logger.info(f"Annotating {len(samples)} samples...")
    annotator.transform(dask_bags).compute(scheduler=scheduler)

    logger.info("All annotations written.")


@click.command(
    entry_point_group="bob.bio.config",
    cls=ConfigCommand,
    epilog="""\b
Examples:

  $ bob bio annotate-samples -vvv config.py -a <annotator> -o /tmp/annotations

You have to define ``samples``, ``reader``, and ``make_key`` in python files
(config.py) as in examples.
""",
)
@click.option(
    "--samples",
    entry_point_group="bob.bio.config",
    required=True,
    cls=ResourceOption,
    help="A list of all samples that you want to annotate. They will be passed "
    "as is to the ``reader`` and ``make-key`` functions.",
)
@click.option(
    "--reader",
    required=True,
    cls=ResourceOption,
    help="A function with the signature of ``data = reader(sample)`` which "
    "takes a sample and returns the loaded data. The returned data is given to "
    "the annotator.",
)
@click.option(
    "--make-key",
    required=True,
    cls=ResourceOption,
    help="A function with the signature of ``key = make_key(sample)`` which "
    "takes a sample and returns a unique str identifier for that sample that "
    "will be use to save it in output_dir. ``key`` generally is the relative "
    "path to a sample's file from the dataset's root directory.",
)
@annotate_common_options
@verbosity_option(logger=logger, expose_value=False)
def annotate_samples(
    samples, reader, make_key, annotator, output_dir, dask_client, **kwargs
):
    """Annotates a list of samples.

    This command is very similar to ``bob bio annotate`` except that it works
    without a database interface. You must provide a list of samples as well as
    two functions:

        def reader(sample):
            # Loads data from a sample.
            # for example:
            data = bob.io.base.load(sample)
            # data will be given to the annotator
            return data

        def make_key(sample):
            # Creates a unique str identifier for this sample.
            # for example:
            return str(sample)
    """
    log_parameters(logger, ignore=("samples",))

    # Allows passing of Sample objects as parameters
    annotator = wrap(["sample"], annotator, output_attribute="annotations")

    # Will save the annotations in the `data` fields to a json file
    annotator = wrap(
        bases=["checkpoint"],
        estimator=annotator,
        features_dir=output_dir,
        extension=".json",
        save_func=save_json,
        load_func=load_json,
        sample_attribute="annotations",
    )

    # Allows reception of Dask Bags
    annotator = wrap(["dask"], annotator)

    # Transformer that splits the samples into several Dask Bags
    to_dask_bags = ToDaskBag(npartitions=50)

    if dask_client is not None:
        scheduler = dask_client
    else:
        scheduler = "single-threaded"

    # Converts samples into a list of DelayedSample objects
    samples_obj = [
        DelayedSample(
            load=functools.partial(reader, s),
            key=make_key(s),
        )
        for s in samples
    ]

    # Splits the samples list into bags
    dask_bags = to_dask_bags.transform(samples_obj)

    logger.info(f"Saving annotations in {output_dir}")
    logger.info(f"Annotating {len(samples_obj)} samples...")
    annotator.transform(dask_bags).compute(scheduler=scheduler)

    logger.info("All annotations written.")



"""Prints a detailed list of all resources that are registered, including the modules, where they have been registered."""

import os
import warnings

import click

import bob.bio.base


@click.command()
@click.option(
    "--types",
    "-t",
    type=click.Choice(
        ["database", "annotator", "pipeline", "config", "dask"],
        case_sensitive=False,
    ),
    multiple=True,
    default=["database", "annotator", "pipeline", "config", "dask"],
    help="Select the resource types that should be listed.",
)
@click.option(
    "--details",
    "-d",
    is_flag=True,
    default=False,
    help="Prints the complete configuration for all resources",
)
@click.option(
    "--packages",
    "-p",
    multiple=True,
    help="If given, shows only resources defined in the given package(s)",
)
@click.option(
    "--no-strip-dummy",
    "-s",
    is_flag=True,
    default=False,
    help="If given, the dummy elements (usually used for testing purposes only) are **not** removed from the list.",
)
def resources(types, details, packages, no_strip_dummy):
    """Lists the currently registered configurations for this environment."""

    # TODO This needs to be updated!

    kwargs = {"verbose": details, "packages": packages}
    if no_strip_dummy:
        kwargs["strip"] = []

    if "database" in types:
        print(
            "\nList of registered databases (can be used after the --database "
            "option):"
        )
        print(bob.bio.base.list_resources("database", **kwargs))

    if "annotator" in types:
        print(
            "\nList of registered annotators (can be used after the "
            "--annotator option):"
        )
        print(bob.bio.base.list_resources("annotator", **kwargs))

    if "pipeline" in types:
        print(
            "\nList of registered pipelines (can be used after the --pipeline "
            "option):"
        )
        print(bob.bio.base.list_resources("pipeline", **kwargs))

    if "config" in types:
        print(
            "\nList of registered configs. Configs may contain multiple "
            "resources and they also allow chain loading (see bob.extension "
            "docs on chain loading). Configs are used as arguments to commands "
            "such as simple):"
        )
        print(bob.bio.base.list_resources("config", **kwargs))

    if "dask" in types:
        print("\nList of registered dask clients")
        print(
            bob.bio.base.list_resources(
                "client", package_prefix="dask.", **kwargs
            )
        )

    print()


def databases(command_line_parameters=None):
    warnings.warn(
        "The databases command is deprecated. You should not be using it!"
    )
    import argparse

    database_replacement = "%s/.bob_bio_databases.txt" % os.environ["HOME"]

    parser = argparse.ArgumentParser(
        description="Prints a list of directories for registered databases",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-D",
        "--database-directories-file",
        metavar="FILE",
        default=database_replacement,
        help="The file, where database directories are stored (to avoid changing the database configurations)",
    )

    args = parser.parse_args(command_line_parameters)

    # get registered databases
    databases = bob.bio.base.utils.resources.database_directories(
        replacements=args.database_directories_file
    )

    # print directories for all databases
    for d in sorted(databases):
        print("\n%s:" % d)

        print("Original data: %s" % databases[d][0])
        if len(databases[d]) > 1:
            print("Annotations: %s" % databases[d][1])



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# @author: Tiago de Freitas Pereira <tiago.pereira@idiap.ch>


"""
Set of legacy functionality for the bob.bio.base.database.Database interface.
"""

import logging
import os
import warnings

import bob.io.base

logger = logging.getLogger(__name__)


def convert_names_to_highlevel(names, low_level_names, high_level_names):
    """
    Converts group names from a low level to high level API

    This is useful for example when you want to return ``db.groups()`` for
    the :py:mod:`bob.bio.base`. Your instance of the database should
    already have ``low_level_names`` and ``high_level_names`` initialized.

    """

    if names is None:
        return None
    mapping = dict(zip(low_level_names, high_level_names))
    if isinstance(names, str):
        return mapping.get(names)
    return [mapping[g] for g in names]


def convert_names_to_lowlevel(names, low_level_names, high_level_names):
    """Same as :py:meth:`convert_names_to_highlevel` but on reverse"""

    if names is None:
        return None
    mapping = dict(zip(high_level_names, low_level_names))
    if isinstance(names, str):
        return mapping.get(names)
    return [mapping[g] for g in names]


def file_names(files, directory, extension):
    """file_names(files, directory, extension) -> paths

    Returns the full path of the given File objects.

    Parameters
    ----------
    files : list of ``bob.db.base.File``
        The list of file object to retrieve the file names for.

    directory : str
        The base directory, where the files can be found.

    extension : str
        The file name extension to add to all files.

    Returns
    -------
    paths : list of :obj:`str`
        The paths extracted for the files, in the same order.
    """
    # return the paths of the files, do not remove duplicates
    return [f.make_path(directory, extension) for f in files]


def sort_files(files):
    """Returns a sorted version of the given list of File's (or other structures
    that define an 'id' data member). The files will be sorted according to their
    id, and duplicate entries will be removed.

    Parameters
    ----------
    files : list of ``bob.db.base.File``
        The list of files to be uniquified and sorted.

    Returns
    -------
    sorted : list of ``bob.db.base.File``
        The sorted list of files, with duplicate `BioFile.id`\\s being removed.
    """
    # sort files using their sort function
    sorted_files = sorted(files)
    # remove duplicates
    return [
        f
        for i, f in enumerate(sorted_files)
        if not i or sorted_files[i - 1].id != f.id
    ]


def check_parameters_for_validity(
    parameters, parameter_description, valid_parameters, default_parameters=None
):
    """Checks the given parameters for validity.

    Checks a given parameter is in the set of valid parameters. It also
    assures that the parameters form a tuple or a list.  If parameters is
    'None' or empty, the default_parameters will be returned (if
    default_parameters is omitted, all valid_parameters are returned).

    This function will return a tuple or list of parameters, or raise a
    ValueError.


    Parameters
    ----------
    parameters : str or list of :obj:`str` or None
        The parameters to be checked. Might be a string, a list/tuple of
        strings, or None.

    parameter_description : str
        A short description of the parameter. This will be used to raise an
        exception in case the parameter is not valid.

    valid_parameters : list of :obj:`str`
        A list/tuple of valid values for the parameters.

    default_parameters : list of :obj:`str` or None
        The list/tuple of default parameters that will be returned in case
        parameters is None or empty. If omitted, all valid_parameters are used.

    Returns
    -------
    tuple
        A list or tuple containing the valid parameters.

    Raises
    ------
    ValueError
        If some of the parameters are not valid.

    """

    if not parameters:
        # parameters are not specified, i.e., 'None' or empty lists
        parameters = (
            default_parameters
            if default_parameters is not None
            else valid_parameters
        )

    if not isinstance(parameters, (list, tuple, set)):
        # parameter is just a single element, not a tuple or list -> transform it
        # into a tuple
        parameters = (parameters,)

    # perform the checks
    for parameter in parameters:
        if parameter not in valid_parameters:
            raise ValueError(
                "Invalid %s '%s'. Valid values are %s, or lists/tuples of those"
                % (parameter_description, parameter, valid_parameters)
            )

    # check passed, now return the list/tuple of parameters
    return parameters


def check_parameter_for_validity(
    parameter, parameter_description, valid_parameters, default_parameter=None
):
    """Checks the given parameter for validity

    Ensures a given parameter is in the set of valid parameters. If the
    parameter is ``None`` or empty, the value in ``default_parameter`` will
    be returned, in case it is specified, otherwise a :py:exc:`ValueError`
    will be raised.

    This function will return the parameter after the check tuple or list
    of parameters, or raise a :py:exc:`ValueError`.

    Parameters
    ----------
    parameter : :obj:`str` or :obj:`None`
                    The single parameter to be checked. Might be a string or None.

    parameter_description : str
                    A short description of the parameter. This will be used to raise an
                    exception in case the parameter is not valid.

    valid_parameters : list of :obj:`str`
                    A list/tuple of valid values for the parameters.

    default_parameter : list of :obj:`str`, optional
                    The default parameter that will be returned in case parameter is None or
                    empty. If omitted and parameter is empty, a ValueError is raised.

    Returns
    -------
    str
                    The validated parameter.

    Raises
    ------
    ValueError
                    If the specified parameter is invalid.

    """

    if parameter is None:
        # parameter not specified ...
        if default_parameter is not None:
            # ... -> use default parameter
            parameter = default_parameter
        else:
            # ... -> raise an exception
            raise ValueError(
                "The %s has to be one of %s, it might not be 'None'."
                % (parameter_description, valid_parameters)
            )

    if isinstance(parameter, (list, tuple, set)):
        # the parameter is in a list/tuple ...
        if len(parameter) > 1:
            raise ValueError(
                "The %s has to be one of %s, it might not be more than one "
                "(%s was given)."
                % (parameter_description, valid_parameters, parameter)
            )
        # ... -> we take the first one
        parameter = parameter[0]

    # perform the check
    if parameter not in valid_parameters:
        raise ValueError(
            "The given %s '%s' is not allowed. Please choose one of %s."
            % (parameter_description, parameter, valid_parameters)
        )

    # tests passed -> return the parameter
    return parameter


class File(object):
    """Abstract class that define basic properties of File objects.

    Your file instance should have at least the self.id and self.path
    properties.
    """

    def __init__(self, path, file_id=None, **kwargs):
        """**Constructor Documentation**

        Initialize the File object with the minimum required data.

        Parameters
        ----------
        path : str
            The path to this file, relative to the basic directory.
            If you use an SQL database, this should be the SQL type String.
            Please do not specify any file extensions.

        file_id : object
            The id of the file (various type). Its type depends on your
            implementation. If you use an SQL database, this should be an SQL type
            like Integer or String. If you are using an automatically determined
            file id, you don't need to specify this parameter.

        Raises
        ------
        NotImplementedError
            If self.id is not set and not specified during initialization through
            `file_id`.
        """

        self.path = path
        """A relative path, which includes file name but excludes file extension"""

        # set file id only, when specified
        if file_id:
            self.id = file_id
            """A unique identifier of the file."""
        else:
            # check that the file id at least exists
            if not hasattr(self, "id"):
                raise NotImplementedError(
                    "Please either specify the file id as parameter, or create an "
                    "'id' member variable in the derived class that is automatically "
                    "determined (e.g. by SQLite)"
                )

        super(File, self).__init__(**kwargs)

    def __lt__(self, other):
        """This function defines the order on the File objects. File objects are
        always ordered by their ID, in ascending order."""
        return self.id < other.id

    def __repr__(self):
        """This function describes how to convert a File object into a string."""
        return "<File('%s': '%s')>" % (str(self.id), str(self.path))

    def make_path(self, directory=None, extension=None):
        """Wraps the current path so that a complete path is formed

        Parameters
        ----------
        directory : :obj:`str`, optional
            An optional directory name that will be prefixed to the returned
            result.
        extension : :obj:`str`, optional
            An optional extension that will be suffixed to the returned filename.
            The extension normally includes the leading ``.`` character as in
            ``.jpg`` or ``.hdf5``.

        Returns
        -------
        str
            Returns a string containing the newly generated file path.
        """
        # assure that directory and extension are actually strings
        # create the path
        return str(os.path.join(directory or "", self.path + (extension or "")))

    def save(
        self, data, directory=None, extension=".hdf5", create_directories=True
    ):
        """Saves the input data at the specified location and using the given
        extension. Override it if you need to save differently.

        Parameters
        ----------
        data : object
            The data blob to be saved (normally a :py:class:`numpy.ndarray`).
        directory : :obj:`str`, optional
            If not empty or None, this directory is prefixed to the final
            file destination
        extension : :obj:`str`, optional
            The extension of the filename - this will control the type of
          output and the codec for saving the input blob.
        create_directories : :obj:`bool`, optional
            Whether to create the required directories to save the data.

        """
        # get the path
        path = self.make_path(directory or "", extension or "")
        # use the bob API to save the data
        bob.io.base.save(data, path, create_directories=create_directories)

    def load(self, directory=None, extension=".hdf5"):
        """Loads the data at the specified location and using the given extension.
        Override it if you need to load differently.

        Parameters
        ----------
        directory : :obj:`str`, optional
            If not empty or None, this directory is prefixed to the final
            file destination
        extension : :obj:`str`, optional
            If not empty or None, this extension is suffixed to the final
            file destination

        Returns
        -------
        object
            The loaded data (normally :py:class:`numpy.ndarray`).

        """
        # get the path
        path = self.make_path(directory or "", extension or "")
        return bob.io.base.load(path)


class FileDatabase(object):
    """Low-level File-based Database API to be used within Bob.

    Not all Databases in Bob need to inherit from this class. Use this class
    only if in your database one sample correlates to one actual file.

    Attributes
    ----------
    original_directory : str
        The directory where the raw files are located.
    original_extension : str
        The extension of raw data files, e.g. ``.png``.
    """

    def __init__(self, original_directory, original_extension, **kwargs):
        super(FileDatabase, self).__init__(**kwargs)
        self.original_directory = original_directory
        self.original_extension = original_extension

    def original_file_names(self, files):
        """Returns the full path of the original data of the given File objects.

        Parameters
        ----------
        files : list of ``bob.db.base.File``
            The list of file object to retrieve the original data file names for.

        Returns
        -------
        list of :obj:`str`
            The paths extracted for the files, in the same order.
        """
        if self.original_directory is None:
            logger.warning(
                "self.original_directory was not provided (must not be None)!"
            )
        if self.original_extension is None:
            logger.warning(
                "self.original_extension was not provided (must not be None)!"
            )
        return file_names(
            files, self.original_directory, self.original_extension
        )

    def original_file_name(self, file):
        """This function returns the original file name for the given File
        object.

        Parameters
        ----------
        file
            ``bob.db.base.File`` or a derivative
            The File objects for which the file name should be retrieved

        Returns
        -------
        str
            The original file name for the given ``bob.db.base.File``
            object.

        Raises
        ------
        ValueError
            if the file is not found.
        """
        # check if directory is set
        if not self.original_directory or not self.original_extension:
            logger.warning(
                "The original_directory and/or the original_extension were not"
                " specified in the constructor."
            )
        # extract file name
        file_name = file.make_path(
            self.original_directory, self.original_extension
        )

        if not self.check_existence or os.path.exists(file_name):
            return file_name

        raise ValueError(
            "The file '%s' was not found. Please check the "
            "original directory '%s' and extension '%s'?"
            % (file_name, self.original_directory, self.original_extension)
        )

    # Deprecated Methods below

    def check_parameters_for_validity(
        self,
        parameters,
        parameter_description,
        valid_parameters,
        default_parameters=None,
    ):
        warnings.warn(
            "check_parameters_for_validity is deprecated. Please use "
            "the equivalent function in this file",
            DeprecationWarning,
            stacklevel=2,
        )
        return check_parameters_for_validity(
            parameters,
            parameter_description,
            valid_parameters,
            default_parameters,
        )

    def check_parameter_for_validity(
        self,
        parameter,
        parameter_description,
        valid_parameters,
        default_parameter=None,
    ):
        warnings.warn(
            "check_parameter_for_validity is deprecated. Please use the "
            "equivalent function in this file",
            DeprecationWarning,
            stacklevel=2,
        )
        return check_parameter_for_validity(
            parameter,
            parameter_description,
            valid_parameters,
            default_parameter,
        )

    def convert_names_to_highlevel(
        self, names, low_level_names, high_level_names
    ):
        warnings.warn(
            "convert_names_to_highlevel is deprecated. Please use the "
            "equivalent function in this file",
            DeprecationWarning,
            stacklevel=2,
        )
        return convert_names_to_highlevel(
            names, low_level_names, high_level_names
        )

    def convert_names_to_lowlevel(
        self, names, low_level_names, high_level_names
    ):
        warnings.warn(
            "convert_names_to_lowlevel is deprecated. Please use the "
            "equivalent function in this file",
            DeprecationWarning,
            stacklevel=2,
        )
        return convert_names_to_lowlevel(
            names, low_level_names, high_level_names
        )

    def file_names(self, files, directory, extension):
        warnings.warn(
            "file_names is deprecated. Please use the "
            "equivalent function in this file",
            DeprecationWarning,
            stacklevel=2,
        )
        return file_names(files, directory, extension)

    def sort(self, files):
        warnings.warn(
            "sort is deprecated. Please use " "sort_files in bob.db.base.utils",
            DeprecationWarning,
            stacklevel=2,
        )
        return sort_files(files)


class Database(FileDatabase):
    """This class is deprecated. New databases should use the
    :py:class:`bob.db.base.FileDatabase` class if required"""

    def __init__(
        self, original_directory=None, original_extension=None, **kwargs
    ):
        warnings.warn(
            "The bob.db.base.Database class is deprecated. "
            "Please use bob.db.base.FileDatabase instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super(Database, self).__init__(
            original_directory, original_extension, **kwargs
        )



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# @author: Tiago de Freitas Pereira <tiago.pereira@idiap.ch>

import glob
import itertools
import os
import shutil
import tempfile
import uuid

import numpy as np
import pytest

from click.testing import CliRunner
from h5py import File as HDF5File
from sklearn.pipeline import make_pipeline

from bob.bio.base.algorithm import Distance
from bob.bio.base.pipelines import (
    BioAlgCheckpointWrapper,
    CSVScoreWriter,
    FourColumnsScoreWriter,
    PipelineSimple,
    dask_bio_pipeline,
)
from bob.bio.base.script.pipeline_simple import (
    pipeline_simple as pipeline_simple_cli,
)
from bob.bio.base.wrappers import wrap_bob_legacy
from bob.io.base.testing_utils import assert_click_runner_result
from bob.pipelines import DelayedSample, Sample, SampleSet
from tests.test_transformers import FakeExtractor, FakePreprocessor


class DummyDatabase:
    def __init__(
        self,
        delayed=False,
        n_references=10,
        n_probes=10,
        dim=10,
        one_d=True,
        some_fta=False,
        all_fta=False,
    ):
        self.delayed = delayed
        self.dim = dim
        self.n_references = n_references
        self.n_probes = n_probes
        self.one_d = one_d
        self.gender_choices = ["M", "F"]
        self.metadata_1_choices = ["A", "B", "C"]
        self.contains_fta = some_fta
        self.all_samples_fta = all_fta

    def _create_random_1dsamples(self, n_samples, offset, dim):
        return [
            Sample(
                np.random.rand(dim),
                key=str(uuid.uuid4()),
                annotations=1,
                template_id=str(i),
                subject_id=str(i),
            )
            for i in range(offset, offset + n_samples)
        ]

    def _create_random_2dsamples(self, n_samples, offset, dim):
        return [
            Sample(
                np.random.rand(dim, dim),
                key=str(uuid.uuid4()),
                annotations=1,
                template_id=str(i),
                subject_id=str(i),
            )
            for i in range(offset, offset + n_samples)
        ]

    def _create_random_sample_set(self, n_sample_set=10, n_samples=2, seed=10):
        # Just generate random samples
        np.random.seed(seed)
        sample_set = [
            SampleSet(
                samples=[],
                key=str(i),
                template_id=str(i),
                subject_id=str(i),
                gender=np.random.choice(self.gender_choices),
                metadata_1=np.random.choice(self.metadata_1_choices),
            )
            for i in range(n_sample_set)
        ]

        offset = 0
        for i, s in enumerate(sample_set):
            if self.one_d:
                s.samples = self._create_random_1dsamples(
                    n_samples, offset, self.dim
                )
            else:
                s.samples = self._create_random_2dsamples(
                    n_samples, offset, self.dim
                )
            if self.contains_fta and i % 2:
                for sample in s.samples[::2]:
                    sample.data = None
            if self.all_samples_fta:
                for sample in s.samples:
                    sample.data = None

            offset += n_samples
            pass

        return sample_set

    def background_model_samples(self):
        samples = [
            sset.samples for sset in self._create_random_sample_set(seed=10)
        ]
        return list(itertools.chain(*samples))

    def references(self, group=None):
        return self._create_random_sample_set(
            self.n_references, self.dim, seed=11
        )

    def probes(self, group=None):
        probes = []

        probes = self._create_random_sample_set(
            n_sample_set=10, n_samples=1, seed=12
        )
        for p in probes:
            p.references = [str(r) for r in list(range(self.n_references))]

        return probes

    def zprobes(self, group=None):
        zprobes = []

        zprobes = self._create_random_sample_set(
            n_sample_set=10, n_samples=1, seed=14
        )
        for p in zprobes:
            p.template_id = "z-" + str(p.template_id)
            p.references = [str(r) for r in list(range(self.n_references))]

        return zprobes

    def treferences(self):
        t_sset = self._create_random_sample_set(
            self.n_references, self.dim, seed=15
        )
        for t in t_sset:
            t.template_id = "t_" + str(t.template_id)
        return t_sset

    @property
    def score_all_vs_all(self):
        return True


def _custom_save_function(data, path):
    base_path, ext = path.rsplit(".", 1)
    filename = base_path + "_with_custom_f." + ext
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with HDF5File(filename, "w") as f:
        f["data"] = data


def _custom_load_function(path):
    base_path, ext = path.rsplit(".", 1)
    filename = base_path + "_with_custom_f." + ext
    assert os.path.isfile(filename)
    with HDF5File(filename, "r") as f:
        return f["data"][()]


class DistanceWithTags(Distance):
    def _more_tags(self):
        return {
            "bob_enrolled_save_fn": _custom_save_function,
            "bob_enrolled_load_fn": _custom_load_function,
            "bob_enrolled_extension": ".hdf5",
        }


def _make_transformer(dir_name):
    pipeline = make_pipeline(
        wrap_bob_legacy(
            FakePreprocessor(),
            dir_name,
            transform_extra_arguments=(("annotations", "annotations"),),
        ),
        wrap_bob_legacy(
            FakeExtractor(),
            dir_name,
        ),
    )

    return pipeline


def test_on_memory():
    with tempfile.TemporaryDirectory() as dir_name:

        def run_pipeline(with_dask, score_all_vs_all):
            database = DummyDatabase()

            transformer = _make_transformer(dir_name)

            biometric_algorithm = Distance()

            pipeline_simple = PipelineSimple(
                transformer,
                biometric_algorithm,
                None,
            )

            if with_dask:
                pipeline_simple = dask_bio_pipeline(
                    pipeline_simple, npartitions=2
                )

            scores = pipeline_simple(
                database.background_model_samples(),
                database.references(),
                database.probes(),
                score_all_vs_all=score_all_vs_all,
            )

            if with_dask:
                scores = scores.compute(scheduler="single-threaded")

            assert len(scores) == 10
            for sample_scores in scores:
                assert len(sample_scores) == 10
                for score in sample_scores:
                    assert isinstance(score.data, float)

        run_pipeline(False, True)
        run_pipeline(False, False)  # Testing checkpoint
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)
        run_pipeline(True, True)
        run_pipeline(True, True)  # Testing checkpoint


def test_checkpoint_bioalg_as_transformer():
    with tempfile.TemporaryDirectory() as dir_name:

        def run_pipeline(
            with_dask,
            score_writer=None,
        ):
            database = DummyDatabase()

            transformer = _make_transformer(dir_name)

            biometric_algorithm = BioAlgCheckpointWrapper(
                Distance(), base_dir=dir_name
            )

            pipeline_simple = PipelineSimple(
                transformer, biometric_algorithm, score_writer
            )

            if with_dask:
                pipeline_simple = dask_bio_pipeline(
                    pipeline_simple, npartitions=2
                )

            scores = pipeline_simple(
                database.background_model_samples(),
                database.references(),
                database.probes(),
                score_all_vs_all=database.score_all_vs_all,
            )

            if pipeline_simple.score_writer is None:
                if with_dask:
                    scores = scores.compute(scheduler="single-threaded")

                assert len(scores) == 10
                for sset in scores:
                    if isinstance(sset[0], DelayedSample):
                        for s in sset:
                            assert len(s.data) == 10
                    else:
                        assert len(sset) == 10
            else:
                writed_scores = pipeline_simple.write_scores(scores)
                scores_dev_path = os.path.join(dir_name, "scores-dev")
                concatenated_scores = pipeline_simple.post_process(
                    writed_scores, scores_dev_path
                )

                if with_dask:
                    concatenated_scores = concatenated_scores.compute(
                        scheduler="single-threaded"
                    )

                if isinstance(
                    pipeline_simple.score_writer, FourColumnsScoreWriter
                ):
                    assert len(open(scores_dev_path).readlines()) == 100
                else:
                    assert (
                        len(open(scores_dev_path).readlines()) == 101
                    )  # Counting the header.

        run_pipeline(False)
        run_pipeline(False)  # Checking if the checkpointing works
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        # Writing scores
        run_pipeline(
            False,
            FourColumnsScoreWriter(os.path.join(dir_name, "final_scores")),
        )
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        # Dask
        run_pipeline(True)
        run_pipeline(True)  # Checking if the checkpointing works
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        # Writing scores
        run_pipeline(
            True, FourColumnsScoreWriter(os.path.join(dir_name, "final_scores"))
        )
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        # CSVWriter
        run_pipeline(
            False, CSVScoreWriter(os.path.join(dir_name, "concatenated_scores"))
        )
        run_pipeline(
            False, CSVScoreWriter(os.path.join(dir_name, "concatenated_scores"))
        )  # Checking if the checkpointing works
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        # CSVWriter + Dask
        run_pipeline(
            True, CSVScoreWriter(os.path.join(dir_name, "concatenated_scores"))
        )
        run_pipeline(
            True, CSVScoreWriter(os.path.join(dir_name, "concatenated_scores"))
        )  # Checking if the checkpointing works


def test_checkpoint_bioalg_with_tags():
    with tempfile.TemporaryDirectory() as dir_name:

        def run_pipeline(
            with_dask,
            score_writer=None,
        ):
            database = DummyDatabase()

            transformer = _make_transformer(dir_name)

            biometric_algorithm = BioAlgCheckpointWrapper(
                DistanceWithTags(), base_dir=dir_name
            )

            pipeline_simple = PipelineSimple(
                transformer, biometric_algorithm, score_writer
            )

            if with_dask:
                pipeline_simple = dask_bio_pipeline(
                    pipeline_simple, npartitions=2
                )

            scores = pipeline_simple(
                database.background_model_samples(),
                database.references(),
                database.probes(),
                score_all_vs_all=database.score_all_vs_all,
            )

            written_scores = pipeline_simple.write_scores(scores)
            scores_dev_path = os.path.join(dir_name, "scores-dev")
            concatenated_scores = pipeline_simple.post_process(
                written_scores, scores_dev_path
            )

            if with_dask:
                concatenated_scores = concatenated_scores.compute(
                    scheduler="single-threaded"
                )

            assert os.path.isfile(
                dir_name
                + "/biometric_references/"
                + database.references()[0].key
                + "_with_custom_f.hdf5"
            )

        run_pipeline(False)
        run_pipeline(False)  # Checking if the checkpointing works
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        # Writing scores
        run_pipeline(
            False,
            FourColumnsScoreWriter(os.path.join(dir_name, "final_scores")),
        )
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        # Dask
        run_pipeline(True)
        run_pipeline(True)  # Checking if the checkpointing works
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        # Writing scores
        run_pipeline(
            True, FourColumnsScoreWriter(os.path.join(dir_name, "final_scores"))
        )
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        # CSVWriter
        run_pipeline(
            False, CSVScoreWriter(os.path.join(dir_name, "concatenated_scores"))
        )
        run_pipeline(
            False, CSVScoreWriter(os.path.join(dir_name, "concatenated_scores"))
        )  # Checking if the checkpointing works
        shutil.rmtree(
            dir_name
        )  # Deleting the cache so it runs again from scratch
        os.makedirs(dir_name, exist_ok=True)

        # CSVWriter + Dask
        run_pipeline(
            True, CSVScoreWriter(os.path.join(dir_name, "concatenated_scores"))
        )
        run_pipeline(
            True, CSVScoreWriter(os.path.join(dir_name, "concatenated_scores"))
        )  # Checking if the checkpointing works


def _run_with_failure(score_all_vs_all, sporadic_fail):
    with tempfile.TemporaryDirectory() as dir_name:
        database = DummyDatabase(
            some_fta=sporadic_fail, all_fta=not sporadic_fail
        )

        transformer = _make_transformer(dir_name)

        biometric_algorithm = Distance()

        pipeline_simple = PipelineSimple(
            transformer,
            biometric_algorithm,
            None,
        )

        scores = pipeline_simple(
            database.background_model_samples(),
            database.references(),
            database.probes(),
            score_all_vs_all=score_all_vs_all,
        )

        assert len(scores) == 10
        for sample_scores in scores:
            assert len(sample_scores) == 10
            for score in sample_scores:
                assert isinstance(score.data, float)


def test_database_sporadic_failure():
    _run_with_failure(False, sporadic_fail=True)
    _run_with_failure(True, sporadic_fail=True)


def test_database_full_failure():
    with pytest.raises(NotImplementedError):
        _run_with_failure(False, sporadic_fail=False)


def test_pipeline_simple_passthrough():
    """Ensure that PipelineSimple accepts a passthrough Estimator."""
    passthrough = make_pipeline(None)
    pipeline = PipelineSimple(passthrough, Distance())
    assert isinstance(pipeline, PipelineSimple)

    pipeline_with_passthrough = make_pipeline("passthrough")
    pipeline = PipelineSimple(pipeline_with_passthrough, Distance())
    assert isinstance(pipeline, PipelineSimple)
    db = DummyDatabase()
    scores = pipeline(
        db.background_model_samples(), db.references(), db.probes()
    )
    assert len(scores) == 10
    for sample_scores in scores:
        assert len(sample_scores) == 10
        for score in sample_scores:
            assert isinstance(score.data, float)


def _create_test_config(path):
    with open(path, "w") as f:
        f.write(
            """
from tests.test_pipeline_simple import DummyDatabase, _make_transformer
from bob.bio.base.pipelines import PipelineSimple
from bob.bio.base.algorithm import Distance

database = DummyDatabase()

transformer = _make_transformer(".")

biometric_algorithm = Distance()

pipeline = PipelineSimple(
    transformer,
    biometric_algorithm,
    None,
)
"""
        )


def _test_pipeline_click_cli(
    cli, options, expected_outputs=("results/scores-dev.csv",)
):
    runner = CliRunner()
    with runner.isolated_filesystem():
        _create_test_config("config.py")
        result = runner.invoke(
            cli,
            [
                "-vv",
                "config.py",
            ]
            + options,
        )
        assert_click_runner_result(result)
        # check for expected_output
        output_files = glob.glob("results/**", recursive=True)
        err_msg = "Found only:\n{output_files}\nin output directory given the options: {options}".format(
            output_files="\n".join(output_files), options=options
        )
        for out in expected_outputs:
            assert os.path.isfile(out), err_msg


def test_pipeline_simple_click_cli():
    # open a click isolated environment

    for options in [
        ["--no-dask", "--memory"],
        ["--no-dask"],
        ["--memory"],
        [],
    ]:
        _test_pipeline_click_cli(pipeline_simple_cli, options)



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

import os

import bob.pipelines

from bob.bio.base.extractor import Extractor
from bob.bio.base.preprocessor import Preprocessor
from bob.bio.base.transformers import (
    ExtractorTransformer,
    PreprocessorTransformer,
)
from bob.bio.base.utils import is_argument_available


def wrap_bob_legacy(
    bob_object,
    dir_name,
    fit_extra_arguments=None,
    transform_extra_arguments=None,
    dask_it=False,
    **kwargs,
):
    """
    Wraps either :any:`bob.bio.base.preprocessor.Preprocessor` or
    :any:`bob.bio.base.extractor.Extractor` with
    :any:`sklearn.base.TransformerMixin` and
    :any:`bob.pipelines.wrappers.CheckpointWrapper` and
    :any:`bob.pipelines.wrappers.SampleWrapper`


    Parameters
    ----------

    bob_object: object
        Instance of :any:`bob.bio.base.preprocessor.Preprocessor` or
        :any:`bob.bio.base.extractor.Extractor`

    dir_name: str
        Directory name for the checkpoints

    fit_extra_arguments: [tuple]
        Same behavior as in Check
        :any:`bob.pipelines.wrappers.fit_extra_arguments`

    transform_extra_arguments: [tuple]
        Same behavior as in Check
        :any:`bob.pipelines.wrappers.transform_extra_arguments`

    dask_it: bool
        If True, the transformer will be a dask graph
    """

    if isinstance(bob_object, Preprocessor):
        transformer = wrap_checkpoint_preprocessor(
            bob_object,
            features_dir=os.path.join(dir_name, "preprocessor"),
            **kwargs,
        )
    elif isinstance(bob_object, Extractor):
        transformer = wrap_checkpoint_extractor(
            bob_object,
            features_dir=os.path.join(dir_name, "extractor"),
            model_path=dir_name,
            fit_extra_arguments=fit_extra_arguments,
            transform_extra_arguments=transform_extra_arguments,
            **kwargs,
        )
    else:
        raise ValueError(
            "`bob_object` should be an instance of `Preprocessor`, `Extractor` and `Algorithm`"
        )

    if dask_it:
        transformer = bob.pipelines.wrap(["dask"], transformer)

    return transformer


def wrap_sample_preprocessor(
    preprocessor,
    transform_extra_arguments=(("annotations", "annotations"),),
    **kwargs,
):
    """
    Wraps :any:`bob.bio.base.preprocessor.Preprocessor` with
    :any:`bob.pipelines.wrappers.CheckpointWrapper` and :any:`bob.pipelines.wrappers.SampleWrapper`

    .. warning::
       This wrapper doesn't checkpoint data

    Parameters
    ----------

    preprocessor: :any:`bob.bio.base.preprocessor.Preprocessor`
       Instance of :any:`bob.bio.base.transformers.PreprocessorTransformer` to be wrapped

    transform_extra_arguments: [tuple]
        Same behavior as in Check :any:`bob.pipelines.wrappers.transform_extra_arguments`

    """

    transformer = PreprocessorTransformer(preprocessor)
    return bob.pipelines.wrap(
        ["sample"],
        transformer,
        transform_extra_arguments=transform_extra_arguments,
    )


def wrap_checkpoint_preprocessor(
    preprocessor,
    features_dir=None,
    transform_extra_arguments=(("annotations", "annotations"),),
    load_func=None,
    save_func=None,
    extension=".hdf5",
):
    """
    Wraps :any:`bob.bio.base.preprocessor.Preprocessor` with
    :any:`bob.pipelines.wrappers.CheckpointWrapper` and :any:`bob.pipelines.wrappers.SampleWrapper`

    Parameters
    ----------

    preprocessor: :any:`bob.bio.base.preprocessor.Preprocessor`
       Instance of :any:`bob.bio.base.transformers.PreprocessorTransformer` to be wrapped

    features_dir: str
       Features directory to be checkpointed (see :any:bob.pipelines.CheckpointWrapper`).

    extension : str, optional
        Extension o preprocessed files (see :any:bob.pipelines.CheckpointWrapper`).

    load_func : None, optional
        Function that loads data to be preprocessed.
        The default is :any:`bob.bio.base.preprocessor.Preprocessor.read_data`

    save_func : None, optional
        Function that saves preprocessed data.
        The default is :any:`bob.bio.base.preprocessor.Preprocessor.write_data`

    transform_extra_arguments: [tuple]
        Same behavior as in Check :any:`bob.pipelines.wrappers.transform_extra_arguments`

    """

    transformer = PreprocessorTransformer(preprocessor)
    return bob.pipelines.wrap(
        ["sample", "checkpoint"],
        transformer,
        load_func=load_func or preprocessor.read_data,
        save_func=save_func or preprocessor.write_data,
        features_dir=features_dir,
        transform_extra_arguments=transform_extra_arguments,
        extension=extension,
    )


def _prepare_extractor_sample_args(
    extractor, transform_extra_arguments, fit_extra_arguments
):
    if transform_extra_arguments is None and is_argument_available(
        "metadata", extractor.__call__
    ):
        transform_extra_arguments = (("metadata", "metadata"),)

    if (
        fit_extra_arguments is None
        and extractor.requires_training
        and extractor.split_training_data_by_client
    ):
        fit_extra_arguments = (("y", "subject"),)

    return transform_extra_arguments, fit_extra_arguments


def wrap_sample_extractor(
    extractor,
    fit_extra_arguments=None,
    transform_extra_arguments=None,
    model_path=None,
    **kwargs,
):
    """
    Wraps :any:`bob.bio.base.extractor.Extractor` with
    :any:`bob.pipelines.wrappers.CheckpointWrapper` and :any:`bob.pipelines.wrappers.SampleWrapper`

    Parameters
    ----------

    extractor: :any:`bob.bio.base.extractor.Preprocessor`
       Instance of :any:`bob.bio.base.transformers.ExtractorTransformer` to be wrapped

    transform_extra_arguments: [tuple], optional
        Same behavior as in Check :any:`bob.pipelines.wrappers.transform_extra_arguments`

    model_path: str
        Path to `extractor_file` in :any:`bob.bio.base.extractor.Extractor`

    """

    extractor_file = (
        os.path.join(model_path, "Extractor.hdf5")
        if model_path is not None
        else None
    )

    transformer = ExtractorTransformer(extractor, model_path=extractor_file)

    (
        transform_extra_arguments,
        fit_extra_arguments,
    ) = _prepare_extractor_sample_args(
        extractor, transform_extra_arguments, fit_extra_arguments
    )

    return bob.pipelines.wrap(
        ["sample"],
        transformer,
        transform_extra_arguments=transform_extra_arguments,
        fit_extra_arguments=fit_extra_arguments,
        **kwargs,
    )


def wrap_checkpoint_extractor(
    extractor,
    features_dir=None,
    fit_extra_arguments=None,
    transform_extra_arguments=None,
    load_func=None,
    save_func=None,
    extension=".hdf5",
    model_path=None,
    **kwargs,
):
    """
    Wraps :any:`bob.bio.base.extractor.Extractor` with
    :any:`bob.pipelines.wrappers.CheckpointWrapper` and :any:`bob.pipelines.wrappers.SampleWrapper`

    Parameters
    ----------

    extractor: :any:`bob.bio.base.extractor.Preprocessor`
       Instance of :any:`bob.bio.base.transformers.ExtractorTransformer` to be wrapped

    features_dir: str
       Features directory to be checkpointed (see :any:bob.pipelines.CheckpointWrapper`).

    extension : str, optional
        Extension o preprocessed files (see :any:bob.pipelines.CheckpointWrapper`).

    load_func : None, optional
        Function that loads data to be preprocessed.
        The default is :any:`bob.bio.base.extractor.Extractor.read_feature`

    save_func : None, optional
        Function that saves preprocessed data.
        The default is :any:`bob.bio.base.extractor.Extractor.write_feature`

    fit_extra_arguments: [tuple]
        Same behavior as in Check :any:`bob.pipelines.wrappers.fit_extra_arguments`

    transform_extra_arguments: [tuple], optional
        Same behavior as in Check :any:`bob.pipelines.wrappers.transform_extra_arguments`

    model_path: str
        See :any:`TransformerExtractor`.

    """

    extractor_file = (
        os.path.join(model_path, "Extractor.hdf5")
        if model_path is not None
        else None
    )

    model_file = (
        os.path.join(model_path, "Extractor.pkl")
        if model_path is not None
        else None
    )
    transformer = ExtractorTransformer(extractor, model_path=extractor_file)

    (
        transform_extra_arguments,
        fit_extra_arguments,
    ) = _prepare_extractor_sample_args(
        extractor, transform_extra_arguments, fit_extra_arguments
    )

    return bob.pipelines.wrap(
        ["sample", "checkpoint"],
        transformer,
        load_func=load_func or extractor.read_feature,
        save_func=save_func or extractor.write_feature,
        model_path=model_file,
        features_dir=features_dir,
        transform_extra_arguments=transform_extra_arguments,
        fit_extra_arguments=fit_extra_arguments,
        **kwargs,
    )



# isort: skip_file
from .csv_database import CSVDatabase, FileSampleLoader, AnnotationsLoader
from .file import BioFile
from .file import BioFileSet
from .database import BioDatabase
from .database import ZTBioDatabase
from .filelist import FileListBioDatabase
from . import filelist  # noqa: F401
from .atnt import AtntBioDatabase  # noqa: F401


# gets sphinx autodoc done right - don't remove it
def __appropriate__(*args):
    """Says object was actually declared here, and not in the import module.
    Fixing sphinx warnings of not being able to find classes, when path is shortened.
    Parameters:

      *args: An iterable of objects to modify

    Resolves `Sphinx referencing issues
    <https://github.com/sphinx-doc/sphinx/issues/3048>`
    """

    for obj in args:
        obj.__module__ = __name__


__appropriate__(
    BioFile,
    BioFileSet,
    BioDatabase,
    ZTBioDatabase,
    CSVDatabase,
    FileSampleLoader,
    FileListBioDatabase,
    AtntBioDatabase,
    AnnotationsLoader,
)
__all__ = [_ for _ in dir() if not _.startswith("_")]



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :


import csv
import os
import uuid

from bob.pipelines import DelayedSample, SampleSet
from bob.pipelines.sample import SAMPLE_DATA_ATTRS

from .abstract_classes import ScoreWriter


class FourColumnsScoreWriter(ScoreWriter):
    """
    Read and write scores using the four columns format
    :py:func:`bob.bio.base.score.load.four_column`
    """

    def __init__(self, path, extension=".txt", **kwargs):
        super().__init__(path, extension, **kwargs)

    def write(self, probe_sampleset: list[SampleSet]):
        """
        Write scores and returns a :py:class:`bob.pipelines.DelayedSample`
        containing the instruction to open the score file
        """

        def _write(probe_sampleset):
            os.makedirs(self.path, exist_ok=True)
            n_lines = 0
            filename = os.path.join(self.path, str(uuid.uuid4()) + ".txt")
            with open(filename, "w") as f:
                for probe in probe_sampleset:
                    # If it's delayed, load it
                    if isinstance(probe[0], DelayedSample):
                        probe.samples = probe.samples[0].data

                    lines = [
                        "{0} {1} {2} {3}\n".format(
                            biometric_reference.subject_id,
                            probe.subject_id,
                            probe.template_id,
                            biometric_reference.data,
                        )
                        for biometric_reference in probe
                    ]
                    n_lines += len(probe)
                    f.writelines(lines)
            return [filename]

        import dask
        import dask.bag

        if isinstance(probe_sampleset, dask.bag.Bag):
            return probe_sampleset.map_partitions(_write)
        return _write(probe_sampleset)


class CSVScoreWriter(ScoreWriter):
    """
    Read and write scores in CSV format, shipping all metadata with the scores

    Parameters
    ----------

    path: str
        Directory to save the scores

    exclude_list: list
        List of metadata to exclude from the CSV file

    """

    def __init__(
        self,
        path,
        exclude_list=tuple(SAMPLE_DATA_ATTRS)
        + ("key", "references", "annotations"),
        **kwargs,
    ):
        super().__init__(path, **kwargs)
        self.exclude_list = exclude_list

    def write(self, probe_sampleset: list[SampleSet]):
        """
        Write scores and returns a :py:class:`bob.pipelines.DelayedSample` containing
        the instruction to open the score file
        """

        def create_csv_header(probe_sampleset):
            first_biometric_reference = probe_sampleset[0]

            probe_dict = dict(
                (k, f"probe_{k}")
                for k in probe_sampleset.__dict__.keys()
                if not (
                    k in self.exclude_list
                    or k.startswith("_")
                    or k == "template_id"
                )
            )

            bioref_dict = dict(
                (k, f"bio_ref_{k}")
                for k in first_biometric_reference.__dict__.keys()
                if not (k in self.exclude_list or k.startswith("_"))
            )

            header = [
                "probe_template_id",
                *probe_dict.values(),
                *bioref_dict.values(),
                "score",
            ]
            return header, probe_dict, bioref_dict

        os.makedirs(self.path, exist_ok=True)
        header, probe_dict, bioref_dict = create_csv_header(probe_sampleset[0])

        filename = os.path.join(self.path, str(uuid.uuid4()))
        with open(filename, "w") as f:
            csv_writer = csv.writer(f)
            rows = []
            for i, probe in enumerate(probe_sampleset):
                # Writing the header
                if i == 0:
                    csv_writer.writerow(header)

                probe_row = [str(probe.template_id)] + [
                    str(getattr(probe, k)) for k in probe_dict
                ]

                # Iterating over the biometric references
                for biometric_reference in probe:
                    bio_ref_row = [
                        str(getattr(biometric_reference, k))
                        for k in list(bioref_dict.keys()) + ["data"]
                    ]

                    rows.append(probe_row + bio_ref_row)

            csv_writer.writerows(rows)
        return [filename]

    def post_process(self, score_paths, path):
        """
        Removing the HEADER of all files
        but the first
        """

        def _post_process(score_paths, path):
            post_processed_scores = []
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "w") as f:
                for i, score in enumerate(score_paths):
                    post_processed_scores.append(score)

                    # Not memory intensive score writing
                    with open(score, "r") as f1:
                        if i > 0:
                            f1.readline()  # skip header line
                        for line in f1:
                            f.write(line)
                    os.remove(score)

            return post_processed_scores

        import dask
        import dask.bag

        if isinstance(score_paths, dask.bag.Bag):
            all_paths = dask.delayed(list)(score_paths)
            return dask.delayed(_post_process)(all_paths, path)
        return _post_process(score_paths, path)
